import logging
import os.path as op
import re
import shutil
from datetime import datetime, timedelta

import numpy as np
from pyrocko.guts import Bool, List

from .plugin import Plugin, PluginConfig


def load_params(filename):
    rc = re.compile(r"([\w]*)\s*([\w.+-]*)")
    params = {}

    with open(filename, mode="r") as par:
        for line in par:
            parsed = rc.match(line)
            if parsed is None:
                continue

            groups = parsed.groups()
            params[groups[0]] = groups[1].strip()

    return params


class GACOSGrid(object):
    def __init__(self, filename, time, data, ulLat, ulLon, dLat, dLon):
        self.filename = filename
        self.time = time

        self.data = data
        self.rows, self.cols = data.shape

        self.dLat = dLat
        self.dLon = dLon

        self.ulLat = ulLat
        self.ulLon = ulLon

        self.llLat = self.ulLat + self.rows * self.dLat
        self.llLon = ulLon

        self.urLon = ulLon + self.cols * self.dLon

    def contains(self, llLat, llLon, dLat, dLon, rows, cols):
        ulLat = llLat + dLat * rows
        urLon = llLon + dLon * cols

        boundary_exception = AssertionError(
            "GACOS Grid does not contain scene!\n"
            " llLat: %.4f urLat: %.4f\n"
            " llLon: %.4f urLon: %.4f\n"
            "Scene:\n"
            " llLat: %.4f urLat: %.4f\n"
            " llLon: %.4f urLon: %.4f"
            % (
                self.llLat,
                self.ulLat,
                self.llLon,
                self.urLon,
                llLat,
                ulLat,
                llLon,
                urLon,
            )
        )

        assert llLat >= self.llLat, boundary_exception
        assert llLon >= self.llLon, boundary_exception

        assert urLon <= self.urLon, boundary_exception
        assert ulLat <= self.ulLat, boundary_exception

    def get_corrections(self, llLat, llLon, dLat, dLon, rows, cols):
        self.contains(llLat, llLon, dLat, dLon, rows, cols)

        ulLat = llLat + dLat * rows
        ulLon = llLon
        urLon = llLon + dLon * cols  # noqa

        row_offset = (self.ulLat - ulLat) // -self.dLat
        col_offset = (ulLon - self.llLon) // self.dLon

        idx_rows = row_offset + (np.arange(rows) * dLat // -self.dLat)
        idx_cols = col_offset + (np.arange(cols) * dLon // self.dLon)

        idx_rows = np.repeat(idx_rows, cols).astype(np.intp)
        idx_cols = np.tile(idx_cols, rows).astype(np.intp)

        return np.flipud(self.data[idx_rows, idx_cols].reshape(rows, cols))

    @classmethod
    def load(cls, filename):
        rsc_file = filename + ".rsc"
        if not op.exists(filename) and not op.exists(rsc_file):
            raise FileNotFoundError(
                "Could not find %s or .rsc file %s" % (filename, rsc_file)
            )

        params = load_params(rsc_file)

        time = datetime.strptime(op.basename(filename)[:8], "%Y%m%d")
        hour = timedelta(hours=float(params["TIME_OF_DAY"].rstrip("UTC")))
        time += hour

        rows = int(params["FILE_LENGTH"])
        cols = int(params["WIDTH"])

        ulLat = float(params["Y_FIRST"])
        ulLon = float(params["X_FIRST"])

        dLat = float(params["Y_STEP"])
        dLon = float(params["X_STEP"])

        data = np.memmap(filename, dtype=np.float32, mode="r", shape=(rows, cols))

        return cls(filename, time, data, ulLat, ulLon, dLat, dLon)


class GACOSConfig(PluginConfig):
    grd_filenames = List.T(default=[], help="List of *two* GACOS gridfiles.")
    applied = Bool.T(default=False, help="Is the correction applied.")


class GACOSCorrection(Plugin):
    def __init__(self, scene=None, config=None):
        self.config = config or GACOSConfig()
        self.scene = scene

        if scene:
            self._log = scene._log.getChild("GACOSCorrection")
        else:
            self._log = logging.getLogger("GACOSCorrection")

        self.grids = []

        assert len(self.config.grd_filenames) <= 2
        for filename in self.config.grd_filenames:
            self.load(filename)

    def has_data(self):
        if len(self.grids) == 2:
            return True
        return False

    def _scene_extent(self):
        frame = self.scene.frame
        return {
            "llLat": frame.llLat,
            "llLon": frame.llLon,
            "dLat": frame.dNdegree,
            "dLon": frame.dEdegree,
            "cols": frame.cols,
            "rows": frame.rows,
        }

    def load(self, filename):
        if len(self.grids) == 2:
            raise AttributeError("already loaded two GACOS grids!")

        if filename.startswith("./"):
            abs_path = op.join(op.dirname(self.scene.meta.filename), filename)
        else:
            abs_path = filename
        if not op.exists(abs_path):
            raise OSError("cannot find GACOS grid %s" % abs_path)

        self._log.info("Loading %s", filename)

        grd = GACOSGrid.load(abs_path)
        grd.contains(**self._scene_extent())

        self.grids.append(grd)
        self.grids = sorted(self.grids, key=lambda grd: grd.time)

        if filename not in self.config.grd_filenames:
            self.config.grd_filenames.append(filename)

    def unload(self):
        self.grids = []
        self.config.grd_filenames = []

    def save(self, dirname):
        for grd_file in self.config.grd_filenames:
            self._log.info("copying GACOS grid %s", grd_file)
            grd_file = op.join(op.dirname(self.scene.meta.filename), grd_file)
            try:
                shutil.copy(grd_file, dirname)
                shutil.copy(grd_file + ".rsc", dirname)
            except shutil.SameFileError:
                pass

        self.config.grd_filenames = [
            "./%s" % op.basename(grd_file) for grd_file in self.config.grd_filenames
        ]

    def get_correction(self):
        if len(self.grids) != 2:
            raise AttributeError("need two GACOS grids to calculate the corrections!")

        extent = self._scene_extent()

        corr_date1 = self.grids[0].get_corrections(**extent)
        corr_date2 = self.grids[1].get_corrections(**extent)

        return corr_date2 - corr_date1

    def apply(self, displacement):
        self._log.info("Applying GACOS model to displacement")
        correction = self.get_correction()
        correction *= 1.0 / np.sin(self.scene.phi)

        displacement -= correction
        return displacement



from hashlib import sha1

from pyrocko.guts import Bool, Object


class PluginConfig(Object):

    applied = Bool.T(default=False)

    def get_hash(self):
        sha = sha1()
        sha.update(str(self).encode())
        return sha.hexdigest()


class Plugin(object):
    def __init__(self, scene, config=None):
        self.scene = scene
        self.config = config or PluginConfig()

    def get_state_hash(self):
        return self.config.get_hash()

    def set_enabled(self, enable):
        assert isinstance(enable, bool)
        self.config.applied = enable
        self.update()

    def is_enabled(self):
        return self.config.applied

    def update(self):
        return self.scene.evChanged()

    def apply(self, displacement):
        raise NotImplementedError
        return displacement



import unittest

import numpy as np

from kite import SandboxScene
from kite.sources import EllipsoidSource, PointCompoundSource
from kite.sources import compound_engine as cm  # noqa

from . import common

km = 1e3
plot = False

benchmark = common.Benchmark()
common.setLogLevel("DEBUG")


class CompoundModelsTest(unittest.TestCase):
    def test_ECM(self):
        nrows = 500
        ncols = 500

        x0 = 250.0
        y0 = 250.0
        depth = 30.0

        rotx = 0.0
        roty = 0.0
        rotz = 0.0

        ax = 1.0
        ay = 1.0
        az = 0.25

        # ax = 1.
        # ay = 1.
        # az = 1.

        P = 1e6
        mu = 0.33e11
        lamda = 0.33e11

        X, Y = np.meshgrid(np.arange(nrows), np.arange(ncols))

        coords = np.empty((nrows * ncols, 2))
        coords[:, 0] = X.ravel()
        coords[:, 1] = Y.ravel()

        # Testing shapeTensor sub-functions
        nu = lamda / (lamda + mu) / 2
        cm.shapeTensor(ax, ay, az, nu)

        @benchmark
        def runECM():
            return cm.ECM(
                coords, x0, y0, depth, rotx, roty, rotz, ax, ay, az, P, mu, lamda
            )

        ue, un, uv, _, _ = runECM()

        self._plot_displacement(un.reshape(nrows, ncols))

    def _test_ECM_against_Octave(self):
        # Lost reference file
        from os import path as p

        from scipy import io

        X, Y = np.meshgrid(np.linspace(-7.0, 7.0, 701), np.linspace(-5.0, 5.0, 501))
        x0 = 0.5
        y0 = -0.25
        depth = 2.75

        rotx = 5.0
        roty = -8.0
        rotz = 30

        ax = 1.0
        ay = 0.75
        az = 0.25

        P = 1e6
        mu = 0.33e11
        lamda = 0.33e11

        coords = np.empty((X.size, 2))
        coords[:, 0] = X.ravel()
        coords[:, 1] = Y.ravel()

        @benchmark
        def runECM():
            return cm.ECM(
                coords, x0, y0, depth, rotx, roty, rotz, ax, ay, az, P, mu, lamda
            )

        ue, un, uv, _, _ = runECM()

        ue = ue.reshape(*X.shape)
        un = un.reshape(*X.shape)
        uv = uv.reshape(*X.shape)

        mat = io.loadmat(
            p.join(p.dirname(__file__), "data", "displacement_ellipsoid_octave.mat")
        )

        np.testing.assert_equal(X, mat["X"])
        np.testing.assert_equal(Y, mat["Y"])

        for pym, comp in zip([ue, un, uv], ["ue", "un", "uv"]):
            m = mat[comp]
            # print([pym.min(), pym.max()], [m.min(), m.max()])
            np.testing.assert_allclose(pym, m, rtol=1e-11)

        self._plot_displacement(uv)
        self._plot_displacement(mat["uv"])

    def testEllipsoidSource(self):
        def r(lo, hi):
            return np.random.randint(lo, high=hi, size=1).astype(float)

        ms = SandboxScene()
        src = EllipsoidSource(
            easting=r(0.0, ms.frame.E.max()),
            northing=r(0.0, ms.frame.N.max()),
            depth=1e3,
        )
        src.regularize()

        ms.addSource(src)

        self._plot_modelScene(ms)

    def _test_pointCDM_against_Octave(self):
        from os import path as p

        from scipy import io

        X, Y = np.meshgrid(np.linspace(-7.0, 7.0, 701), np.linspace(-5.0, 5.0, 501))
        x0 = 0.5
        y0 = -0.25
        depth = 2.75

        rotx = 5.0
        roty = -8.0
        rotz = 30.0

        dVx = 0.00144
        dVy = 0.00128
        dVz = 0.00072

        nu = 0.25

        coords = np.empty((X.size, 2))
        coords[:, 0] = X.ravel()
        coords[:, 1] = Y.ravel()

        @benchmark
        def run_pointCDM():
            return cm.pointCDM(
                coords, x0, y0, depth, rotx, roty, rotz, dVx, dVy, dVz, nu
            )

        ue, un, uv = run_pointCDM()

        ue = ue.reshape(*X.shape)
        un = un.reshape(*X.shape)
        uv = uv.reshape(*X.shape)

        mat = io.loadmat(
            p.join(p.dirname(__file__), "data", "displacement_pcdm_octave.mat")
        )

        np.testing.assert_equal(X, mat["X"])
        np.testing.assert_equal(Y, mat["Y"])

        for pym, comp in zip([ue, un, uv], ["ue", "un", "uv"]):
            m = mat[comp]
            # print([pym.min(), pym.max()], [m.min(), m.max()])
            np.testing.assert_allclose(pym, m, rtol=1e-9)

        self._plot_displacement(mat["uv"])
        self._plot_displacement(uv)

    def testPointCompoundSourceSource(self):
        def r(lo, hi):
            return np.random.randint(lo, high=hi, size=1).astype(float)

        ms = SandboxScene()
        src = PointCompoundSource(
            easting=r(0.0, ms.frame.E.max()),
            northing=r(0.0, ms.frame.N.max()),
            depth=1e3,
        )
        src.regularize()
        ms.addSource(src)

        self._plot_modelScene(ms)

    @staticmethod
    def _plot_modelScene(ms):
        if not plot:
            ms.down
            return

        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.gca()
        ms.processSources()

        ax.imshow(
            np.flipud(ms.down),
            aspect="equal",
            extent=[0, ms.frame.E.max(), 0, ms.frame.N.max()],
        )
        plt.show()

    @staticmethod
    def _plot_displacement(u):
        if not plot:
            u
            return

        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.gca()

        ax.imshow(u)
        plt.show()


if __name__ == "__main__":
    plot = True
    unittest.main(exit=False)
    print(benchmark)



from os import path

import numpy as np
import pyqtgraph as pg
import pyqtgraph.parametertree.parameterTypes as pTypes
from PyQt5 import QtCore
from pyqtgraph import dockarea

from kite.qt_utils import _viridis_data
from kite.util import calcPrecision, formatScalar

__all__ = ["KiteView", "KitePlot", "KiteToolColormap", "KiteParameterGroup"]


d2r = np.pi / 180.0


def get_resource(filename):
    return path.join(path.dirname(path.realpath(__file__)), "res", filename)


class KiteView(dockarea.DockArea):
    title = "View Prototype"

    def __init__(self):
        dockarea.DockArea.__init__(self)
        self.tool_docks = []
        self.tools = getattr(self, "tools", {})

        self.colormap = KiteToolColormap(self.main_widget)

        dock_main = dockarea.Dock(
            self.title, autoOrientation=False, widget=self.main_widget
        )
        dock_colormap = dockarea.Dock(
            "Colormap", autoOrientation=False, widget=self.colormap
        )
        dock_colormap.setStretch(1, None)

        for i, (name, tool) in enumerate(self.tools.items()):
            self.tool_docks.append(
                dockarea.Dock(name, widget=tool, size=(2, 2), autoOrientation=False)
            )
            self.addDock(self.tool_docks[-1], position="bottom")

        self.addDock(dock_main, position="left")
        self.addDock(dock_colormap, position="right")

    @QtCore.pyqtSlot()
    def activateView(self):
        pass

    @QtCore.pyqtSlot()
    def deactivateView(self):
        pass


class KitePlotThread(QtCore.QThread):
    completeSignal = QtCore.pyqtSignal()

    def __init__(self, plot):
        super().__init__()
        self.plot = plot

    def run(self):
        plot = self.plot
        plot._data = plot.components_available[plot.component][1](plot.model)
        self.completeSignal.emit()


class LOSArrow(pg.GraphicsWidget, pg.GraphicsWidgetAnchor):
    def __init__(self, model):
        pg.GraphicsWidget.__init__(self)
        pg.GraphicsWidgetAnchor.__init__(self)

        self.model = model

        self.arrow = pg.ArrowItem(
            parent=self,
            angle=0.0,
            brush=(0, 0, 0, 180),
            pen=(255, 255, 255),
            pxMode=True,
        )

        self.label = pg.LabelItem(
            "Towards Sat.", justify="right", size="8pt", parent=self
        )
        self.label.anchor(itemPos=(1.0, -1.0), parentPos=(1.0, 0.0))
        # self.label.setBrush(pg.mkBrush(255, 255, 255, 180))
        # self.label.setFont(QtGui.QFont(
        #     "Helvetica", weight=QtGui.QFont.DemiBold))

        self.orientArrow()
        self.model.sigSceneChanged.connect(self.orientArrow)
        self.setFlag(self.ItemIgnoresTransformations)

    @QtCore.pyqtSlot()
    def orientArrow(self):
        phi = np.nanmedian(self.model.scene.phi)
        theta = np.nanmedian(self.model.scene.theta)

        angle = 180.0 - np.rad2deg(phi)

        theta_f = theta / (np.pi / 2)

        tipAngle = 30.0 + theta_f * 20.0
        tailLen = 15 + theta_f * 15.0

        self.arrow.setStyle(
            angle=0.0, tipAngle=tipAngle, tailLen=tailLen, tailWidth=6, headLen=25
        )
        self.arrow.setRotation(angle)

        rect_label = self.label.boundingRect()
        rect_arr = self.arrow.boundingRect()

        self.label.setPos(-rect_label.width() / 2.0, rect_label.height() * 1.33)

    def setParentItem(self, parent):
        pg.GraphicsWidget.setParentItem(self, parent)

    def boundingRect(self):
        return QtCore.QRectF(0, 0, self.width(), self.height())


class KitePlot(pg.PlotWidget):
    def __init__(self, model, los_arrow=False, auto_downsample=False):
        pg.PlotWidget.__init__(self)
        self.model = model
        self.draw_time = 0.0
        self._data = None
        self.updateThread = KitePlotThread(self)
        self.updateThread.completeSignal.connect(self._updateImageFromData)

        border_pen = pg.mkPen(255, 255, 255, 50)
        self.image = pg.ImageItem(
            None, autoDownsample=auto_downsample, border=border_pen, useOpenGL=True
        )

        self.hillshade = None
        self.elevation = None

        self.setAspectLocked(True)
        self.plotItem.getAxis("left").setZValue(100)
        self.plotItem.getAxis("bottom").setZValue(100)

        self.hint = {
            "east": 0.0,
            "north": 0.0,
            "value": np.nan,
            "measure": self.component.title(),
            "vlength": "03",
            "precision": "3",
        }

        self.hint_text = pg.LabelItem(
            text="", justify="right", size="8pt", parent=self.plotItem
        )
        self.hint_text.anchor(itemPos=(1.0, 0.0), parentPos=(1.0, 0.0))
        self.hint_text.setOpacity(0.6)

        if self.model.frame.isMeter():
            self.hint_text.text_template = (
                '<span style="font-family: monospace; color: #fff;'
                'background-color: #000;">'
                "East {east:08.2f} m | North {north:08.2f} m |"
                " {measure} {value:{length}.{precision}f}</span>"
            )
            self.setLabels(bottom=("Easting", "m"), left=("Northing", "m"))
        elif self.model.frame.isDegree():
            self.hint_text.text_template = (
                '<span style="font-family: monospace; color: #fff;'
                'background-color: #000;">'
                "Lon {east:03.3f}&deg; | Lat {north:02.3f}&deg; |"
                " {measure} {value:{length}.{precision}f}</span>"
            )
            self.setLabels(bottom="Longitude", left="Latitude")

        self.addItem(self.image)
        self.update()

        self.transFromFrame()
        self._move_sig = pg.SignalProxy(
            self.image.scene().sigMouseMoved, rateLimit=25, slot=self.mouseMoved
        )

        if los_arrow:
            self.addLOSArrow()

        # self.addIsocurve()
        # self.scalebar()

    def addLOSArrow(self):
        self.los_arrow = LOSArrow(self.model)
        self.los_arrow.setParentItem(self.graphicsItem())
        self.los_arrow.anchor(
            itemPos=(1.0, 0.0), parentPos=(1, 0.0), offset=(-10.0, 40.0)
        )

    def enableHillshade(self):
        from scipy.signal import fftconvolve

        frame = self.model.frame
        scene = self.model.scene

        elevation = scene.get_elevation()

        contrast = 1.0
        elevation_angle = 45.0
        azimuth = 45.0

        size_ramp = 10
        ramp = np.linspace(-1, 1, size_ramp)[::-1]

        ramp_x = np.tile(ramp, size_ramp).reshape(size_ramp, size_ramp)
        ramp_y = ramp_x.T
        ramp = ramp_x + ramp_y

        ramp = ramp / ramp.max() * contrast
        # ramp = np.array([[-1, -.5, 0], [0, .5, 1.]]) * contrast
        # convolution of two 2-dimensional arrays

        shad = fftconvolve(elevation, ramp, mode="valid")
        shad = -1.0 * shad

        # if there are strong artificial edges in the data, shades get
        # dominated by them. Cutting off the largest and smallest 2% of
        # shades helps

        percentile2 = np.quantile(shad, 0.02)
        percentile98 = np.quantile(shad, 0.98)

        shad = np.where(shad > percentile98, percentile98, shad)
        shad = np.where(shad < percentile2, percentile2, shad)
        shad -= shad.min()
        shad /= shad.max()

        # normalize to range [0 1]
        hillshade = pg.ImageItem(
            None, autoDownsample=False, border=None, useOpenGL=False
        )
        hillshade.resetTransform()
        hillshade.scale(frame.dE, frame.dN)

        elev_img = pg.ImageItem(
            None, autoDownsample=False, border=None, useOpenGL=False
        )
        elev_img.resetTransform()
        elev_img.scale(frame.dE, frame.dN)

        elev_img.updateImage(elevation.T, autoLevels=True)
        hillshade.updateImage(shad.T, autoLevels=True)

        self.hillshade = hillshade
        self.elevation = elev_img
        self.removeItem(self.image)
        self.addItem(self.elevation)
        self.addItem(self.hillshade)
        self.addItem(self.image)

        self.update()

    def disableHillshade(self):
        if not self.hillshade:
            return
        self.removeItem(self.hillshade)
        self.hillshade = None

    def transFromFrame(self):
        frame = self.model.frame

        self.image.resetTransform()
        self.image.scale(frame.dE, frame.dN)

    def scalebar(self):
        """Not working"""
        self.scale_bar = pg.ScaleBar(10, width=5, suffix="m")
        self.scale_bar.setParentItem(self.plotItem)
        self.scale_bar.anchor((1, 1), (1, 1), offset=(-20, -20))

    @property
    def component(self):
        return self._component

    @component.setter
    def component(self, component):
        if component not in self.components_available.keys():
            raise AttributeError("Invalid component %s" % component)
        self._component = component
        self.update()

    @property
    def data(self):
        if self._data is None:
            self._data = self.components_available[self.component][1](self.model)
        return self._data
        # return self._data  # np.nan_to_num(_data)

    @QtCore.pyqtSlot()
    def update(self, obj=None):
        if not self.updateThread.isRunning():
            self.updateThread.start()

    @QtCore.pyqtSlot()
    def _updateImageFromData(self):
        self.image.updateImage(self.data.T, opacity=0.7 if self.hillshade else 1.0)

        self.hint["precision"], self.hint["vlength"] = calcPrecision(self.data)
        self.mouseMoved()

    @QtCore.pyqtSlot(object)
    def mouseMoved(self, event=None):
        frame = self.model.frame
        if event is None:
            return
        elif self.image.sceneBoundingRect().contains(event[0]):
            map_pos = self.plotItem.vb.mapSceneToView(event[0])
            if not map_pos.isNull():
                img_pos = self.image.mapFromScene(*event)
                value = self.image.image[int(img_pos.x()), int(img_pos.y())]

                self.hint["east"] = map_pos.x()
                self.hint["north"] = map_pos.y()
                self.hint["value"] = value

                if frame.isDegree():
                    self.hint["east"] += frame.llLon
                    self.hint["north"] += frame.llLat

        self.hint["length"] = (
            "03" if np.isnan(self.hint["value"]) else self.hint["vlength"]
        )
        self.hint_text.setText(self.hint_text.text_template.format(**self.hint))


class KiteToolColormap(pg.HistogramLUTWidget):
    def __init__(self, plot):
        pg.HistogramLUTWidget.__init__(self, image=plot.image)
        self._plot = plot
        self.prev_levels = None
        self.symmetric_colormap = True

        zero_marker = pg.InfiniteLine(pos=0, angle=0, pen="w", movable=False)
        zero_marker.setValue(0.0)
        zero_marker.setZValue(1000)
        self.vb.addItem(zero_marker)

        self.axis.setLabel("Displacement / m")
        # self.plot.rotate(-90)
        # self.layout.rotate(90)
        # self.gradient.setOrientation('bottom')
        self.setSymColormap()
        self._plot.image.sigImageChanged.connect(self.imageChanged)

        self.sigLevelsChanged.connect(self.symmetricLevels)
        # self.isoCurveControl()

    def set_symmetric_colormap(self, symmetric):
        self.symmetric_colormap = symmetric

    @QtCore.pyqtSlot()
    def imageChanged(self):
        if self._plot.component == "weight":
            self.setQualitativeColormap()
        else:
            self.setSymColormap()

        max_range = np.nanmax(np.abs(self._plot.data))
        if max_range is not np.nan:
            self.vb.setYRange(-max_range, max_range)

    @QtCore.pyqtSlot(object)
    def symmetricLevels(self, *args):
        if not self.symmetric_colormap:
            return

        levels = self.getLevels()
        if self.prev_levels is None:
            self.prev_levels = levels

        min_changed = bool(self.prev_levels[0] - levels[0])
        func = min if min_changed else max

        max_lvl = abs(func(levels))
        self.prev_levels = levels
        self.setLevels(-max_lvl, max_lvl)

    def setSymColormap(self):
        cmap = {
            "ticks": [
                [0.0, (0, 0, 0, 255)],
                [1e-3, (106, 0, 31, 255)],
                [0.5, (255, 255, 255, 255)],
                [1.0, (8, 54, 104, 255)],
            ],
            "mode": "rgb",
        }
        cmap = {
            "ticks": [
                [0.0, (0, 0, 0)],
                [1e-3, (172, 56, 56)],
                [0.5, (255, 255, 255)],
                [1.0, (51, 53, 120)],
            ],
            "mode": "rgb",
        }

        relevant_data = np.abs(self._plot.data[np.isfinite(self._plot.data)])
        if np.any(relevant_data):
            lvl_max = np.quantile(relevant_data, 0.999)
        else:
            lvl_max = 1.0

        self.gradient.restoreState(cmap)
        self.setLevels(-lvl_max, lvl_max)

    def setQualitativeColormap(self):
        nc = len(_viridis_data) - 1
        cmap = {"mode": "rgb"}
        cmap["ticks"] = [[float(i) / nc, c] for i, c in enumerate(_viridis_data)]
        self.gradient.restoreState(cmap)
        self.setLevels(np.nanmin(self._plot.data), np.nanmax(self._plot.data))

    def isoCurveControl(self):
        iso_ctrl = pg.InfiniteLine(pos=0, angle=0, pen="g", movable=True)
        iso_ctrl.setValue(0.0)
        iso_ctrl.setZValue(1000)

        def isolineChange():
            self._plot.iso.setLevel(iso_ctrl.value())

        iso_ctrl.sigDragged.connect(isolineChange)
        self.vb.addItem(iso_ctrl)


class KiteParameterGroup(pTypes.GroupParameter):
    def __init__(self, model, model_attr=None, **kwargs):
        self.model = model
        self.model_attr = model_attr

        self.parameters = getattr(self, "parameters", [])

        if isinstance(self.parameters, list):
            self.parameters = dict.fromkeys(self.parameters)

        pTypes.GroupParameter.__init__(self, **kwargs)
        self.updateValues()

    def updateValues(self):
        if not hasattr(self, "parameters"):
            return

        if self.model_attr is not None and hasattr(self.model, self.model_attr):
            model = getattr(self.model, self.model_attr)
        else:
            model = self.model

        for param, f in self.parameters.items():
            QtCore.QCoreApplication.processEvents()
            try:
                if callable(f):
                    value = f(model)
                    # print('Updating %s: %s (func)' % (param, value))
                else:
                    value = getattr(model, param)
                    # print('Updating %s: %s (getattr)' % (param, value))
                try:
                    value = formatScalar(float(value))
                except ValueError:
                    pass
            except AttributeError:
                value = "n/a"
            try:
                self.child(param).setValue(value)
            except Exception:
                self.addChild(
                    {"name": param, "value": value, "type": "str", "readonly": True}
                )

    def pushChild(self, child, **kwargs):
        self.insertChild(0, child, **kwargs)



from __future__ import absolute_import

import shutil
import tempfile
import unittest
from os import path as op

import numpy as np

from kite import Scene

from . import common

# format (dl_dir, load_file)
filenames = {
    "matlab": ("myanmar_alos_dsc_ionocorr.mat", None),
    "gmtsar": ("gmtsar/", "gmtsar/unwrap_ll.los_ll.grd"),
    "roi_pac": ("roi_pac/", "roi_pac/geo_20160113-20160206_atmo_2rlks_c10_cut.unw"),
    "gamma": ("gamma/", "gamma/asc"),
    "isce": ("isce/", "isce/filt_topophase.unw.geo"),
}

common.setLogLevel("DEBUG")


class SceneIOTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tmp_dir = tempfile.mkdtemp(prefix="kite")

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmp_dir)


def _create_test_func(fmt, dl_path, filename):
    def f(self):
        common.get_test_data(dl_path)

        if filename is None:
            load_path = dl_path
        else:
            load_path = filename
        load_path = op.join(common.data_dir, load_path)

        fn_save = op.join(self.tmp_dir, "kite-%s" % fmt)

        sc1 = Scene.import_data(load_path)
        sc1.save(fn_save)

        sc2 = Scene.load(fn_save)

        np.testing.assert_equal(sc1.displacement, sc2.displacement)
        np.testing.assert_equal(sc1.phi, sc2.phi)
        np.testing.assert_equal(sc1.theta, sc2.theta)

    f.__name__ = "test_import_%s" % fmt

    return f


for fmt, fns in filenames.items():
    setattr(SceneIOTest, "test_import_%s" % fmt, _create_test_func(fmt, *fns))



import math
import time
from collections import OrderedDict

import pyqtgraph as pg
import pyqtgraph.parametertree.parameterTypes as pTypes
from PyQt5 import QtCore, QtGui, QtWidgets

from kite.qt_utils import SliderWidgetParameterItem
from kite.quadtree import QuadtreeConfig

from .base import KiteParameterGroup, KitePlot, KiteView


class KiteQuadtree(KiteView):
    title = "Scene.quadtree"

    def __init__(self, spool):
        model = spool.model
        self.model = model
        self.main_widget = KiteQuadtreePlot(model)
        self.tools = {}

        self.param_quadtree = KiteParamQuadtree(model, self.main_widget, expanded=True)
        self.parameters = [self.param_quadtree]

        model.sigSceneModelChanged.connect(self.modelChanged)

        KiteView.__init__(self)

    def modelChanged(self):
        self.main_widget.update()
        self.main_widget.transFromFrame()
        self.main_widget.updateFocalPoints()

        self.param_quadtree.updateValues()
        self.param_quadtree.onConfigUpdate()
        self.param_quadtree.updateEpsilonLimits()

    @QtCore.pyqtSlot()
    def activateView(self):
        self.model.scene.quadtree.ensureTree()
        self.main_widget.activatePlot()

    @QtCore.pyqtSlot()
    def deactivateView(self):
        self.main_widget.deactivatePlot()


class QQuadLeaf(QtCore.QRectF):

    leaf_outline = pg.mkPen((255, 255, 255), width=1)
    leaf_fill = pg.mkBrush(0, 0, 0, 0)

    def __init__(self, leaf):
        self.id = leaf.id
        QtCore.QRectF.__init__(
            self,
            QtCore.QPointF(leaf.llE, leaf.llN + leaf.sizeN),
            QtCore.QPointF(leaf.llE + leaf.sizeE, leaf.llN),
        )

    def getRectItem(self):
        item = QtWidgets.QGraphicsRectItem(self)
        item.setPen(self.leaf_outline)
        item.setBrush(self.leaf_fill)
        item.setZValue(1e8)
        return item


class QQuadSelectedLeaf(QQuadLeaf):
    leaf_outline = pg.mkPen((202, 60, 60), width=1)
    leaf_fill = pg.mkBrush(202, 60, 60, 120)


class KiteQuadtreePlot(KitePlot):
    def __init__(self, model):

        self.components_available = {
            "mean": [
                "Node.mean displacement",
                lambda sp: sp.quadtree.leaf_matrix_means,
            ],
            "median": [
                "Node.median displacement",
                lambda sp: sp.quadtree.leaf_matrix_medians,
            ],
            "weight": [
                "Node.weight covariance",
                lambda sp: sp.quadtree.leaf_matrix_weights,
            ],
        }

        self._component = "median"

        KitePlot.__init__(self, model=model, los_arrow=True)

        focalpoint_color = (78, 255, 0)
        focalpoint_outline_color = (0, 0, 0)
        self.focal_points = pg.ScatterPlotItem(
            size=3.5,
            pen=pg.mkPen(focalpoint_outline_color, width=0.3),
            brush=pg.mkBrush(focalpoint_color),
            antialias=True,
        )

        self.setMenuEnabled(True)

        self.highlighted_leaves = []
        self.selected_leaves = []
        self.outlined_leaves = None

        self.eraseBox = QtWidgets.QGraphicsRectItem(0, 0, 1, 1)
        self.eraseBox.setPen(pg.mkPen((202, 60, 60), width=1, style=QtCore.Qt.DotLine))
        self.eraseBox.setBrush(pg.mkBrush(202, 60, 60, 40))
        self.eraseBox.setZValue(1e9)
        self.eraseBox.hide()
        self.addItem(self.eraseBox, ignoreBounds=True)

        self.vb = self.getViewBox()
        self.vb.mouseDragEvent = self.mouseDragEvent
        self.vb.keyPressEvent = self.blacklistSelectedLeaves

        self.addItem(self.focal_points)

        # self.model.sigFrameChanged.connect(self.transFromFrame)
        # self.model.sigFrameChanged.connect(self.transFromFrameScatter)

    @QtCore.pyqtSlot()
    def covarianceChanged(self):
        if self._component == "weight":
            self.update()

    def activatePlot(self):
        self.model.sigQuadtreeChanged.connect(self.unselectLeaves)
        self.model.sigQuadtreeChanged.connect(self.update)
        self.model.sigQuadtreeChanged.connect(self.updateFocalPoints)
        self.model.sigQuadtreeChanged.connect(self.updateLeavesOutline)
        self.model.sigCovarianceChanged.connect(self.covarianceChanged)

        self.update()
        self.updateLeavesOutline()
        self.updateFocalPoints()

    def deactivatePlot(self):
        self.model.sigQuadtreeChanged.disconnect(self.unselectLeaves)
        self.model.sigQuadtreeChanged.disconnect(self.update)
        self.model.sigQuadtreeChanged.disconnect(self.updateFocalPoints)
        self.model.sigQuadtreeChanged.disconnect(self.updateLeavesOutline)
        self.model.sigCovarianceChanged.disconnect(self.covarianceChanged)

    def transFromFrameScatter(self):
        self.focal_points.resetTransform()
        self.focal_points.scale(self.model.frame.dE, self.model.frame.dN)

    @QtCore.pyqtSlot()
    def updateFocalPoints(self):
        if self.model.quadtree.leaf_focal_points.size == 0:
            self.focal_points.clear()
        else:
            self.focal_points.setData(
                pos=self.model.quadtree.leaf_focal_points, pxMode=True
            )

    def updateEraseBox(self, p1, p2):
        r = QtCore.QRectF(p1, p2)
        r = self.vb.childGroup.mapRectFromParent(r)
        self.eraseBox.r = r
        self.eraseBox.setPos(r.topLeft())
        self.eraseBox.resetTransform()
        self.eraseBox.scale(r.width(), r.height())
        self.eraseBox.show()

    @QtCore.pyqtSlot(object)
    def mouseDragEvent(self, ev, axis=None):
        if ev.button() & (QtCore.Qt.LeftButton | QtCore.Qt.MidButton):
            return pg.ViewBox.mouseDragEvent(self.vb, ev, axis)

        ev.accept()
        if ev.isFinish():
            self.eraseBox.hide()
            self.selectLeaves()
        else:
            self.updateEraseBox(ev.buttonDownPos(), ev.pos())

    def getQLeaves(self, cls):
        return [cls(lf) for lf in self.model.quadtree.leaves]

    @QtCore.pyqtSlot()
    def selectLeaves(self):
        self.unselectLeaves()

        self.selected_leaves = [
            lf
            for lf in self.getQLeaves(QQuadSelectedLeaf)
            if self.eraseBox.r.contains(lf)
        ]
        for lf in self.selected_leaves:
            leaf_item = lf.getRectItem()
            leaf_item.setZValue(1e9)
            leaf_item.setToolTip("Press Del to remove")

            self.highlighted_leaves.append(leaf_item)
            self.addItem(leaf_item)

    @QtCore.pyqtSlot()
    def unselectLeaves(self):
        if self.selected_leaves:
            for lf in self.highlighted_leaves:
                self.removeItem(lf)
            del self.highlighted_leaves
            del self.selected_leaves
            self.highlighted_leaves = []
            self.selected_leaves = []

    @QtCore.pyqtSlot(object)
    def blacklistSelectedLeaves(self, ev):
        if ev.key() == QtCore.Qt.Key_Delete:
            self.model.quadtree.blacklistLeaves(lf.id for lf in self.selected_leaves)

    @QtCore.pyqtSlot()
    def updateLeavesOutline(self):
        group = QtWidgets.QGraphicsItemGroup()
        for lf in self.model.quadtree.leaves:
            group.addToGroup(QQuadLeaf(lf).getRectItem())
        group.setOpacity(0.4)

        if self.outlined_leaves is not None:
            self.vb.removeItem(self.outlined_leaves)

        self.outlined_leaves = group
        self.vb.addItem(self.outlined_leaves)


class KiteParamQuadtree(KiteParameterGroup):
    sigEpsilon = QtCore.pyqtSignal(float)
    sigNanFraction = QtCore.pyqtSignal(float)
    sigTileMaximum = QtCore.pyqtSignal(float)
    sigTileMinimum = QtCore.pyqtSignal(float)

    def __init__(self, model, plot, *args, **kwargs):
        self.plot = plot
        self.sig_guard = True
        self.sp = model

        kwargs["type"] = "group"
        kwargs["name"] = "Scene.quadtree"
        self.parameters = OrderedDict(
            [
                ("nleaves", None),
                ("reduction_rms", None),
                # ('reduction_efficiency', None),
                # ('epsilon_min', None),
                ("nnodes", None),
            ]
        )

        KiteParameterGroup.__init__(self, model=model, model_attr="quadtree", **kwargs)

        model.sigQuadtreeConfigChanged.connect(self.onConfigUpdate)
        model.sigQuadtreeChanged.connect(self.updateValues)

        self.sigEpsilon.connect(model.qtproxy.setEpsilon)
        self.sigNanFraction.connect(model.qtproxy.setNanFraction)
        self.sigTileMaximum.connect(model.qtproxy.setTileMaximum)
        self.sigTileMinimum.connect(model.qtproxy.setTileMinimum)

        def updateGuard(func):
            def wrapper(*args, **kwargs):
                if not self.sig_guard:
                    func()

            return wrapper

        # Epsilon control
        @updateGuard
        @QtCore.pyqtSlot()
        def updateEpsilon():
            self.sigEpsilon.emit(self.epsilon.value())

        eps_decimals = -math.floor(math.log10(model.quadtree.epsilon_min or 1e-3))
        eps_steps = round(
            (model.quadtree.epsilon - model.quadtree.epsilon_min) * 0.1, 3
        )

        p = {
            "name": "epsilon",
            "type": "float",
            "value": model.quadtree.epsilon,
            "default": model.quadtree._epsilon_init,
            "step": eps_steps,
            "limits": (model.quadtree.epsilon_min, 3 * model.quadtree._epsilon_init),
            "editable": True,
            "decimals": eps_decimals,
            "slider_exponent": 2,
            "tip": QuadtreeConfig.epsilon.help,
        }
        self.epsilon = pTypes.SimpleParameter(**p)
        self.epsilon.itemClass = SliderWidgetParameterItem
        self.epsilon.sigValueChanged.connect(updateEpsilon)

        @updateGuard
        @QtCore.pyqtSlot()
        def updateNanFrac():
            self.sigNanFraction.emit(self.nan_allowed.value())

        p = {
            "name": "nan_allowed",
            "type": "float",
            "value": model.quadtree.nan_allowed,
            "default": QuadtreeConfig.nan_allowed.default(),
            "step": 0.05,
            "limits": (0.0, 1.0),
            "editable": True,
            "decimals": 2,
            "tip": QuadtreeConfig.nan_allowed.help,
        }
        self.nan_allowed = pTypes.SimpleParameter(**p)
        self.nan_allowed.itemClass = SliderWidgetParameterItem
        self.nan_allowed.sigValueChanged.connect(updateNanFrac)

        # Tile size controls
        @updateGuard
        @QtCore.pyqtSlot()
        def updateTileSizeMin():
            self.sigTileMinimum.emit(self.tile_size_min.value())

        frame = model.frame
        max_px = max(frame.shape)
        max_pxd = max(frame.dE, frame.dN)
        limits = (max_pxd * model.quadtree.min_node_length_px, max_pxd * (max_px / 4))

        tile_decimals = 2

        p = {
            "name": "tile_size_min",
            "type": "float",
            "value": model.quadtree.tile_size_min,
            "default": QuadtreeConfig.tile_size_min.default(),
            "limits": limits,
            "step": 250,
            "slider_exponent": 2,
            "editable": True,
            "suffix": " m" if frame.isMeter() else "°",
            "decimals": 0 if frame.isMeter() else tile_decimals,
            "tip": QuadtreeConfig.tile_size_min.help,
        }
        self.tile_size_min = pTypes.SimpleParameter(**p)
        self.tile_size_min.itemClass = SliderWidgetParameterItem

        @updateGuard
        def updateTileSizeMax():
            self.sigTileMaximum.emit(self.tile_size_max.value())

        p.update(
            {
                "name": "tile_size_max",
                "value": model.quadtree.tile_size_max,
                "default": QuadtreeConfig.tile_size_max.default(),
                "tip": QuadtreeConfig.tile_size_max.help,
            }
        )
        self.tile_size_max = pTypes.SimpleParameter(**p)
        self.tile_size_max.itemClass = SliderWidgetParameterItem

        self.tile_size_min.sigValueChanged.connect(updateTileSizeMin)
        self.tile_size_max.sigValueChanged.connect(updateTileSizeMax)

        # Component control
        @QtCore.pyqtSlot()
        def changeComponent():
            self.plot.component = self.components.value()

        p = {
            "name": "display",
            "values": {
                "QuadNode.mean": "mean",
                "QuadNode.median": "median",
                "QuadNode.weight": "weight",
            },
            "value": "mean",
            "tip": "Change displayed component",
        }
        self.components = pTypes.ListParameter(**p)
        self.components.sigValueChanged.connect(changeComponent)

        @QtCore.pyqtSlot()
        def changeCorrection():
            model.quadtree.setCorrection(correction_method.value())
            self.updateEpsilonLimits()

        p = {
            "name": "setCorrection",
            "values": {
                "Mean (Jonsson, 2002)": "mean",
                "Median (Jonsson, 2002)": "median",
                "Bilinear (Jonsson, 2002)": "bilinear",
                "SD (Jonsson, 2002)": "std",
            },
            "value": model.quadtree.config.correction,
            "tip": QuadtreeConfig.correction.help,
        }
        correction_method = pTypes.ListParameter(**p)
        correction_method.sigValueChanged.connect(changeCorrection)

        self.sig_guard = False
        self.pushChild(correction_method)
        self.pushChild(self.tile_size_max)
        self.pushChild(self.tile_size_min)
        self.pushChild(self.nan_allowed)
        self.pushChild(self.epsilon)
        self.pushChild(self.components)

    @QtCore.pyqtSlot()
    def onConfigUpdate(self):
        self.sig_guard = True
        for p in ["epsilon", "nan_allowed", "tile_size_min", "tile_size_max"]:
            param = getattr(self, p)
            param.setValue(getattr(self.sp.quadtree, p))
        self.sig_guard = False

    def updateEpsilonLimits(self):
        self.epsilon.setLimits(
            (self.sp.quadtree.epsilon_min, 3 * self.sp.quadtree._epsilon_init)
        )



from kite import SandboxScene
from kite.sources import EllipsoidSource

km = 1e3

sandbox = SandboxScene()

ellipsoid_source = EllipsoidSource(
    northing=40 * km,
    easting=40 * km,
    depth=4 * km,
    length_x=100,
    length_y=200,
    length_z=350,
    roation_x=42.0,
)
sandbox.addSource(ellipsoid_source)

sandbox.save("/tmp/sandbox_scene.yml")
playground = SandboxScene.load("/tmp/sandbox_scene.yml")



#!/usr/bin/env python3
import unittest

import matplotlib.pyplot as plt
import numpy as np

from kite import Scene

from . import common

benchmark = common.Benchmark()
common.setLogLevel("DEBUG")


class TestCovariance(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        file = common.get_test_data("myanmar_alos_dsc_ionocorr.mat")
        cls.sc = Scene.import_data(file)

    def test_covariance(self):
        cov = self.sc.covariance
        cov.epsilon = 0.02
        cov.subsampling = 24

        d = []
        d.append(("Full", cov._calcCovarianceMatrix(method="full", nthreads=0)))
        d.append(("Focal", cov._calcCovarianceMatrix(method="focal")))

        for _, c1 in d:
            for _, c2 in d:
                np.testing.assert_allclose(c1, c2, rtol=200, atol=2e3, verbose=True)

    def test_synthetic_noise(self):
        self.sc.covariance.syntheticNoise()
        self.sc.covariance.variance

    def test_quadtree_noise(self):
        rstate = np.random.RandomState()
        self.sc.covariance.getQuadtreeNoise(rstate=rstate)

    def test_covariance_parallel(self):
        self.sc.quadtree.epsilon = 0.07
        self.sc.quadtree.tile_size_max = 11000

        cov = self.sc.covariance
        cov.config.adaptive_subsampling = True

        @benchmark
        def calc_exp():
            return cov._calcCovarianceMatrix(method="full", nthreads=0)

        @benchmark
        def calc_exp_cos():
            cov.setModelFunction("exponential_cosine")
            return cov._calcCovarianceMatrix(method="full", nthreads=0)

    @benchmark
    def _test_covariance_single_thread(self):
        cov = self.sc.covariance
        cov._calcCovarianceMatrix(method="full", nthreads=1)

    @benchmark
    def test_covariance_focal(self):
        cov = self.sc.covariance
        cov._calcCovarianceMatrix(method="focal")

    @unittest.skip("Skip!")
    def _test_covariance_visual(self):
        cov = self.sc.covariance
        cov.epsilon = 0.02
        cov.subsampling = 10
        # l = self.sc.quadtree.leaves[0]
        d = []
        d.append(("Full", cov._calcCovarianceMatrix(method="full", nthreads=0)))
        d.append(("Focal", cov._calcCovarianceMatrix(method="focal")))

        fig, _ = plt.subplots(1, len(d))
        for i, (title, mat) in enumerate(d):
            print("%s Max %f" % ((title, np.nanmax(mat)), mat.shape))
            fig.axes[i].imshow(mat)
            fig.axes[i].set_title(title)
        plt.show()

    def test_covariance_spatial(self):
        cov = self.sc.covariance
        quad = self.sc.quadtree  # noqa

        @benchmark
        def calc(c):
            cov, dist = c.covariance_spatial
            # assert np.all(np.isfinite(cov))

        for i in range(10):
            calc(cov)

        print(benchmark)


if __name__ == "__main__":
    unittest.main(exit=False)
    print(benchmark)



import numpy as np

from kite import util


def test_trim_matrix():
    arr = np.full((100, 100), np.nan)

    arr[-1, -1] = 1.0
    assert util.trimMatrix(arr).shape == (1, 1)

    arr[-2, -2] = 1.0
    assert util.trimMatrix(arr).shape == (2, 2)

    arr[np.diag_indices_from(arr)] = 1.0
    assert util.trimMatrix(arr).shape == arr.shape



import numpy as np
import pyrocko.orthodrome as od
from pyrocko.guts import Float

from . import compound_engine as ce
from .base import SandboxSource, SourceProcessor

d2r = np.pi / 180.0
r2d = 180.0 / np.pi
km = 1e3


class EllipsoidSource(SandboxSource):

    __implements__ = "CompoundModel"

    length_x = Float.T(
        help="Length of x semi-axis of ellisoid before rotation in [m]", default=1.0
    )
    length_y = Float.T(
        help="Length of y semi-axis of ellisoid before rotation in [m]", default=1.0
    )
    length_z = Float.T(
        help="Length of z semi-axis of ellisoid before rotation in [m]", default=1.0
    )
    rotation_x = Float.T(
        help="Clockwise rotation of ellipsoid around x-axis in [deg]", default=0.0
    )
    rotation_y = Float.T(
        help="Clockwise rotation of ellipsoid around y-axis in [deg]", default=0.0
    )
    rotation_z = Float.T(
        help="Clockwise rotation of ellipsoid around z-axis in [deg]", default=0.0
    )

    mu = Float.T(
        help="Shear modulus, 2. Lame constant in [GPa]",
        default=8.0,
    )
    lamda = Float.T(help="Lame constant in [GPa]", default=8.0)
    cavity_pressure = Float.T(help="Pressure on the cavity walls in [GPa]", default=0.5)

    @property
    def volume(self):
        K = self.lamda + 2 * self.mu / 3  # Bulk Modulus
        V = 4.0 / 3 * np.pi * self.length_x * self.length_y * self.length_z
        V = (self.cavity_pressure * V) / K
        return V

    def ECMParameters(self):
        if self._sandbox:
            east_shift, north_shift = od.latlon_to_ne_numpy(
                self._sandbox.frame.llLat, self._sandbox.frame.llLon, self.lat, self.lon
            )
        else:
            east_shift = 0.0
            north_shift = 0.0

        params = {
            "x0": self.easting + east_shift,
            "y0": self.northing + north_shift,
            "z0": self.depth,
            "rotx": self.rotation_x,
            "roty": self.rotation_y,
            "rotz": self.rotation_z,
            "ax": self.length_x,
            "ay": self.length_y,
            "az": self.length_z,
            "P": self.cavity_pressure * 1e9,
            "mu": self.mu * 1e9,
            "lamda": self.lamda * 1e9,
        }
        return params


class PointCompoundSource(SandboxSource):

    __implements__ = "CompoundModel"

    rotation_x = Float.T(
        help="Clockwise rotation of ellipsoid around x-axis in [deg]", default=0.0
    )
    rotation_y = Float.T(
        help="Clockwise rotation of ellipsoid around y-axis in [deg]", default=0.0
    )
    rotation_z = Float.T(
        help="Clockwise rotation of ellipsoid around z-axis in [deg]", default=0.0
    )
    dVx = Float.T(help="Volume change in x-plane in [m3]", default=1.0)
    dVy = Float.T(help="Volume change in y-plane in [m3]", default=1.0)
    dVz = Float.T(help="Volume change in z-plane in [m3]", default=1.0)
    nu = Float.T(help="Poisson's ratio, typically 0.25", default=0.25)

    @property
    def volume(self):
        # After Nikkhoo, M et al. 2017
        return (self.dVx + self.dVy + self.dVz) / 1.8

    def pointCDMParameters(self):
        if self._sandbox:
            east_shift, north_shift = od.latlon_to_ne_numpy(
                self._sandbox.frame.llLat, self._sandbox.frame.llLon, self.lat, self.lon
            )
        else:
            east_shift = 0.0
            north_shift = 0.0

        params = {
            "x0": self.easting + east_shift,
            "y0": self.northing + north_shift,
            "z0": self.depth,
            "rotx": self.rotation_x,
            "roty": self.rotation_y,
            "rotz": self.rotation_z,
            "dVx": self.dVx**3,
            "dVy": self.dVy**3,
            "dVz": self.dVz**3,
            "nu": self.nu,
        }
        return params


class CompoundModelProcessor(SourceProcessor):

    __implements__ = "CompoundModel"

    def process(self, sources, sandbox, nthreads=0):
        result = {
            "processor_profile": dict(),
            "displacement.e": np.zeros(sandbox.frame.npixel),
            "displacement.n": np.zeros(sandbox.frame.npixel),
            "displacement.d": np.zeros(sandbox.frame.npixel),
        }

        coords = sandbox.frame.coordinatesMeter

        for src in sources:
            if isinstance(src, EllipsoidSource):
                res = ce.ECM(coords, **src.ECMParameters())
            elif isinstance(src, PointCompoundSource):
                res = ce.pointCDM(coords, **src.pointCDMParameters())
            else:
                raise AttributeError("Source of wrong type!")
            result["displacement.e"] += res[0]
            result["displacement.n"] += res[1]
            result["displacement.d"] += res[2]

        return result



import logging
import os.path as op
import re

import numpy as np
import pyrocko.orthodrome as od
import shapefile
import utm
from scipy import stats

from kite.scene import Scene, SceneConfig

log = logging.getLogger("bbd2kite")

d2r = np.pi / 180.0
r2d = 180.0 / np.pi


class DataStruct(dict):
    def __getattr__(self, attr):
        return self[attr]

    def __setattr__(self, attr, value):
        self[attr] = value


def read_projection(filename):
    prj_filename = op.splitext(filename)[0] + ".prj"
    if not op.exists(prj_filename):
        log.warning("Could not find %s, defaulting to UTM Zone 32N")
        return 32, "N"

    with open(prj_filename, "r") as f:
        matches = re.findall(r'[\'"](\w*)[\'"]', f.read())

    if not matches:
        log.warning("Could not find projection in %s", prj_filename)
        return 32, "N"

    projection = matches[0]
    if "UTM" not in projection:
        raise AttributeError("Projection is not UTM: %s" % projection)

    zone, letter = int(projection[-3:-1]), projection[-1]
    return zone, letter


def read_shapefile(filename):
    log.info("Loading data from %s", filename)
    shp = shapefile.Reader(filename)

    npoints = shp.numRecords
    field_name_map = {fld[0].lower(): fld[0] for fld in shp.fields}

    data = DataStruct()
    data.bbox = shp.bbox

    data.ps_mean_v = np.zeros(npoints)
    data.ps_mean_var = np.zeros(npoints)
    los_n = np.zeros(npoints)
    los_e = np.zeros(npoints)
    los_u = np.zeros(npoints)

    coords = np.zeros((npoints, 2))

    for isr, sr in enumerate(shp.iterShapeRecords()):
        shape = sr.shape
        record = sr.record
        # assert shape.shapeType == 11

        los_n[isr] = getattr(record, field_name_map["los_north"])
        los_e[isr] = getattr(record, field_name_map["los_east"])
        los_u[isr] = -getattr(record, field_name_map["los_up"])

        data.ps_mean_v[isr] = getattr(record, field_name_map["mean_velo"])
        data.ps_mean_var[isr] = getattr(record, field_name_map["var_mean_v"])

        coords[isr] = shape.points[0]

    data.phi = np.arctan2(los_n, los_e)
    data.theta = np.arcsin(los_u)

    data.easts = coords[:, 0]
    data.norths = coords[:, 1]

    return data


def bin_ps_data(data, bins=(800, 800)):
    log.debug("Binning mean velocity data...")
    bin_vels, edg_E, edg_N, _ = stats.binned_statistic_2d(
        data.easts, data.norths, data.ps_mean_v, statistic="mean", bins=bins
    )

    log.debug("Binning LOS angles...")
    bin_phi, _, _, _ = stats.binned_statistic_2d(
        data.easts, data.norths, data.phi, statistic="mean", bins=bins
    )
    bin_theta, _, _, _ = stats.binned_statistic_2d(
        data.easts, data.norths, data.theta, statistic="mean", bins=bins
    )

    log.debug("Binning mean velocity variance...")
    bin_mean_var, _, _, _ = stats.binned_statistic_2d(
        data.easts, data.norths, data.ps_mean_var, statistic="mean", bins=bins
    )

    data.bin_mean_var = bin_mean_var
    data.bin_ps_mean_v = bin_vels

    data.bin_phi = bin_phi
    data.bin_theta = bin_theta

    data.bin_edg_N = edg_N
    data.bin_edg_E = edg_E

    return data


def bbd2kite(filename, px_size=(500, 500), import_var=False, convert_m=True):
    """Convert BGR BodenBewegungsdienst PS velocity data to a Kite Scene

    Loads the mean PS velocities (from e.g. ``ps_plot(..., -1)``) from a
    BGR BodenBewegungsdienst, and grids the data into mean velocity bins.
    The LOS velocities will be converted to a Kite Scene
    (:class:`~kite.Scene`).

    :param filename: Name of the BGR BBD as ESRI shapefile.
    :type filename: str
    :param px_size: Size of pixels in North and East in meters.
        Default (500, 500).
    :type px_size: tuple
    :param convert_m: Convert displacement to meters, default True.
    :type convert_m: bool
    :param import_var: Import the mean velocity variance, this information
        is used by the Kite scene to define the covariance.
    :param import_var: bool
    """
    data = read_shapefile(filename)

    if convert_m:
        data.ps_mean_v /= 1e3
        data.ps_mean_var /= 1e3

    # lengthN = od.distance_accurate50m(
    #     data.bbox[1], data.bbox[0],
    #     data.bbox[3], data.bbox[0])
    # lengthE = od.distance_accurate50m(
    #     data.bbox[1], data.bbox[0],
    #     data.bbox[1], data.bbox[2])

    lengthE = data.bbox[2] - data.bbox[0]
    lengthN = data.bbox[3] - data.bbox[1]

    bins = (round(lengthE / px_size[0]), round(lengthN / px_size[1]))

    bin_ps_data(data, bins=bins)

    log.debug("Setting up the Kite Scene")
    config = SceneConfig()
    zone, letter = read_projection(filename)

    llLat, llLon = utm.to_latlon(data.bbox[0], data.bbox[1], zone, letter)
    config.frame.llLat = llLat
    config.frame.llLon = llLon

    config.frame.dE = data.bin_edg_E[1] - data.bin_edg_E[0]
    config.frame.dN = data.bin_edg_N[1] - data.bin_edg_N[0]
    config.frame.spacing = "meter"

    scene_name = op.basename(op.abspath(filename))
    config.meta.scene_title = "%s (BodenbewegungsDienst import)" % scene_name
    config.meta.scene_id = scene_name
    # config.meta.time_master = data.tmin.timestamp()
    # config.meta.time_slave = data.tmax.timestamp()

    scene = Scene(
        theta=data.bin_theta,
        phi=data.bin_phi,
        displacement=data.bin_ps_mean_v,
        config=config,
    )

    if import_var:
        scene.displacement_px_var = data.bin_mean_var

    return scene


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="""Convert BodenbewegungsDienst PS displacements into a Kite scene.

Loads the PS velocities delivered by BGR BodenbewegungsDienst
(https://bodenbewegungsdienst.bgr.de) and grids the data in mean velocity bins.
The mean LOS velocities will be converted into a Kite Scene.

The data is delivered in ESRI shapefile format.
""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "file", type=str, default=".", help="BodenbewegungsDienst shape file."
    )
    parser.add_argument(
        "--resolution",
        "-r",
        nargs=2,
        metavar=("mE", "mN"),
        dest="resolution",
        type=int,
        default=(500, 500),
        help="pixel size of the output grid in east and north (meter)."
        " Default is 500 m by 500 m.",
    )
    parser.add_argument(
        "--save",
        "-s",
        default=None,
        type=str,
        dest="save",
        help="filename to save the Kite scene to. If not given, the scene"
        " will be opened in spool GUI.",
    )
    parser.add_argument(
        "--force",
        "-f",
        default=False,
        action="store_true",
        dest="force",
        help="force overwrite of an existing scene.",
    )
    parser.add_argument(
        "--keep-mm",
        action="store_true",
        default=False,
        help="keep mm/a and do not convert to m/a.",
    )
    parser.add_argument(
        "--import-var",
        action="store_true",
        dest="import_var",
        default=False,
        help="import the variance, which is added to Kite's"
        " scene covariance matrix.",
    )
    parser.add_argument(
        "-v",
        action="count",
        default=0,
        help="verbosity, add multiple to increase verbosity.",
    )

    args = parser.parse_args()

    log_level = logging.INFO - args.v * 10
    logging.basicConfig(level=log_level if log_level > 0 else 0)

    fn_save = args.save
    if args.save:
        for fn in (fn_save, fn_save + ".yml", fn_save + ".npz"):
            if op.exists(fn) and not args.force:
                raise UserWarning("File %s exists! Use --force to overwrite." % fn_save)

    scene = bbd2kite(
        filename=args.file,
        px_size=args.resolution,
        import_var=args.import_var,
        convert_m=not args.keep_mm,
    )

    if fn_save is not None:
        fn_save.rstrip(".yml")
        fn_save.rstrip(".npz")

        log.info(
            "Saving BGR BodenBewegunsdienst scene to file %s[.yml/.npz]...",  # noqa
            fn_save,
        )
        scene.save(args.save)

    elif scene:
        scene.spool()


if __name__ == "__main__":
    main()



import logging
import os
from os import path as op

import pyqtgraph.parametertree.parameterTypes as pTypes
from PyQt5 import QtCore, QtGui, QtWidgets, uic
from pyqtgraph.parametertree.parameterTypes import WidgetParameterItem

SCRIPT_DIRECTORY = op.dirname(op.abspath(__file__))

_viridis_data = [
    [68, 1, 84],
    [69, 6, 90],
    [70, 12, 95],
    [71, 18, 101],
    [71, 24, 106],
    [72, 29, 111],
    [72, 34, 115],
    [71, 39, 119],
    [71, 44, 123],
    [70, 49, 126],
    [69, 54, 129],
    [67, 59, 131],
    [66, 64, 133],
    [64, 68, 135],
    [62, 73, 137],
    [60, 77, 138],
    [58, 82, 139],
    [56, 86, 139],
    [54, 90, 140],
    [52, 94, 141],
    [50, 98, 141],
    [49, 102, 141],
    [47, 106, 141],
    [45, 110, 142],
    [44, 114, 142],
    [42, 118, 142],
    [40, 122, 142],
    [39, 125, 142],
    [38, 129, 142],
    [36, 133, 141],
    [35, 137, 141],
    [33, 140, 141],
    [32, 144, 140],
    [31, 148, 139],
    [30, 152, 138],
    [30, 155, 137],
    [30, 159, 136],
    [31, 163, 134],
    [33, 167, 132],
    [36, 170, 130],
    [40, 174, 127],
    [44, 177, 125],
    [50, 181, 122],
    [56, 185, 118],
    [62, 188, 115],
    [69, 191, 111],
    [77, 194, 107],
    [85, 198, 102],
    [94, 201, 97],
    [103, 204, 92],
    [112, 206, 86],
    [121, 209, 81],
    [131, 211, 75],
    [141, 214, 68],
    [151, 216, 62],
    [162, 218, 55],
    [173, 220, 48],
    [183, 221, 41],
    [194, 223, 34],
    [205, 224, 29],
    [215, 226, 25],
    [225, 227, 24],
    [236, 228, 26],
    [246, 230, 31],
]
_viridis_data.reverse()


def validateFilename(filename):
    if not filename:
        return False
    filename = op.abspath(filename)
    filedir = op.dirname(filename)
    if filename == "" or filedir == "":
        return False
    if op.isdir(filename) or not os.access(filedir, os.W_OK):
        QtWidgets.QMessageBox.critical(
            None, "Path Error", f"Could not access file <b>{filename}</b>"
        )
        return False
    return True


def loadUi(uifile, baseinstance=None):
    return uic.loadUi(uifile, baseinstance)


class SliderWidget(QtWidgets.QWidget):
    """
    shows a horizontal/vertical slider with a label showing its value
    """

    sigValueChanged = QtCore.pyqtSignal(object)  # value

    def __init__(
        self, horizontal=True, parent=None, decimals=3, step=0.005, slider_exponent=1
    ):
        QtWidgets.QWidget.__init__(self, parent)
        self.vmin = None
        self.vmax = None
        self.slider_exponent = slider_exponent
        self.decimals = decimals
        self.step = 100
        self.valueLen = 2
        self.suffix = None
        self._value = None

        self.spin = QtWidgets.QDoubleSpinBox()
        self.spin.setDecimals(decimals)
        self.spin.setSingleStep(step)
        self.spin.setFrame(False)
        self.spin.valueChanged.connect(self._spin_updated)

        self.slider = QtWidgets.QSlider(
            QtCore.Qt.Orientation(1 if horizontal else 0), self
        )  # 1 = hor.
        self.slider.setTickPosition(
            QtWidgets.QSlider.TicksAbove if horizontal else QtWidgets.QSlider.TicksLeft
        )
        self.slider.setRange(0, 99)
        self.slider.sliderMoved.connect(self._slider_updated)

        layout = QtWidgets.QHBoxLayout() if horizontal else QtWidgets.QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.slider)
        layout.addWidget(self.spin)

    def value(self):
        return self._value

    def setValue(self, val):
        self.setSliderValue(val)
        self.spin.setValue(val)
        self._value = val

    def setSliderValue(self, val):
        if val is None:
            val = self.vmin

        elif self.vmin is not None and self.vmax is not None:
            if val <= self.vmin:
                val = self.vmin
            val = (val - self.vmin) / (self.vmax - self.vmin)

            val **= 1.0 / self.slider_exponent
            val *= 99
            val = int(round(val))

        self.slider.setValue(val)

    def setRange(self, vmin, vmax):
        """
        vmin, vmax -> arbitrary values that are not equal
        """
        if vmin == vmax:
            raise ValueError("limits must be different values")
        self.vmin = float(min(vmin, vmax))
        self.vmax = float(max(vmin, vmax))

        self.spin.setRange(self.vmin, self.vmax)

    def setOpts(self, bounds=None):
        if bounds is not None:
            self.setRange(*bounds)

    def setSuffix(self, suffix=None):
        self.suffix = suffix
        self.spin.setSuffix(suffix)

    @QtCore.pyqtSlot(int)
    def _slider_updated(self, val):
        val /= 99  # val -> 0...1
        val **= self.slider_exponent

        if self.vmin is not None and self.vmax is not None:
            val = val * (self.vmax - self.vmin) + self.vmin

        val = round(val, self.decimals)

        if self._value != val:
            self._value = val
            self.sigValueChanged.emit(val)

        self.spin.setValue(val)

    @QtCore.pyqtSlot(float)
    def _spin_updated(self, val):
        self.setSliderValue(self._value)

        if self._value != val:
            self._value = val
            self.sigValueChanged.emit(val)


class SliderWidgetParameterItem(WidgetParameterItem):
    """Enabling Slider widget for Parameter"""

    def makeWidget(self):
        opts = self.param.opts

        step = opts.get("step", 0.1)
        decimals = opts.get("decimals", 2)
        slider_exponent = opts.get("slider_exponent", 1)

        w = SliderWidget(decimals=decimals, step=step, slider_exponent=slider_exponent)

        limits = opts.get("limits", None)
        if limits is not None:
            w.setRange(*limits)

        value = opts.get("value", None)
        if value is not None:
            w.setValue(value)

        suffix = opts.get("suffix", None)
        w.setSuffix(suffix)

        self.hideWidget = False
        w.sigChanged = w.sigValueChanged
        w.sigChanging = w.sigValueChanged

        return w


class SceneLogModel(QtCore.QAbstractTableModel, logging.Handler):
    def __init__(self, model, *args, **kwargs):
        QtCore.QAbstractTableModel.__init__(self, *args, **kwargs)
        logging.Handler.__init__(self)

        self.log_records = []
        self.app = None
        self.model = model
        self.model.sigLogRecord.connect(self.newRecord)

    def data(self, idx, role):
        rec = self.log_records[idx.row()]

        if role == QtCore.Qt.DisplayRole:
            if idx.column() == 0:
                return int(rec.levelno)
            elif idx.column() == 1:
                return "%s:%s" % (rec.levelname, rec.name)
            elif idx.column() == 2:
                return rec.getMessage()

        elif role == QtCore.Qt.ItemDataRole:
            return rec

        elif role == QtCore.Qt.ToolTipRole:
            if idx.column() == 0:
                return rec.levelname
            elif idx.column() == 1:
                return "%s.%s" % (rec.module, rec.funcName)
            elif idx.column() == 2:
                return "Line %d" % rec.lineno

    @property
    def nlogs(self):
        return len(self.log_records)

    def rowCount(self, idx):
        return self.nlogs

    def columnCount(self, idx):
        return 3

    @QtCore.pyqtSlot(object)
    def newRecord(self, record):
        self.beginInsertRows(QtCore.QModelIndex(), self.nlogs, self.nlogs)
        self.log_records.append(record)
        self.endInsertRows()


class SceneLog(QtWidgets.QDialog):

    levels = {
        50: "Critical",
        40: "Error",
        30: "Warning",
        20: "Info",
        10: "Debug",
        0: "All",
    }

    class LogEntryDelegate(QtWidgets.QStyledItemDelegate):

        levels = {
            50: QtWidgets.QStyle.SP_MessageBoxCritical,
            40: QtWidgets.QStyle.SP_MessageBoxCritical,
            30: QtWidgets.QStyle.SP_MessageBoxWarning,
            20: QtWidgets.QStyle.SP_MessageBoxInformation,
            10: QtWidgets.QStyle.SP_FileIcon,
        }

        def paint(self, painter, option, idx):
            # paint icon instead of log_lvl
            if idx.column() == 0:
                getIcon = QtWidgets.QApplication.style().standardIcon
                icon = getIcon(self.levels[idx.data()])
                icon.paint(painter, option.rect)
            else:
                QtWidgets.QStyledItemDelegate.paint(self, painter, option, idx)

    class LogFilter(QtCore.QSortFilterProxyModel):
        def __init__(self, *args, **kwargs):
            QtCore.QSortFilterProxyModel.__init__(self, *args, **kwargs)
            self.setLevel(30)

        def setLevel(self, level):
            self.level = level
            self.setFilterRegExp(str(level))
            self.invalidate()

    def __init__(self, app, model):
        QtWidgets.QDialog.__init__(self, parent=app)
        logging_ui = op.join(
            op.dirname(op.realpath(__file__)), "spool", "res", "logging.ui"
        )
        loadUi(logging_ui, baseinstance=self)

        self.move(
            self.parent().window().mapToGlobal(self.parent().window().rect().center())
            - self.mapToGlobal(self.rect().center())
        )

        self.closeButton.setIcon(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogCloseButton)
        )

        self.table_filter = self.LogFilter()
        self.table_filter.setFilterKeyColumn(0)
        self.table_filter.setDynamicSortFilter(True)
        self.table_filter.setSourceModel(model.log)

        model.log.rowsInserted.connect(self.newLogRecord)

        self.tableView.setModel(self.table_filter)
        self.tableView.setItemDelegate(self.LogEntryDelegate())

        self.tableView.setColumnWidth(0, 30)
        self.tableView.setColumnWidth(1, 200)

        self.filterBox.addItems([lvl_name for lvl_name in self.levels.values()])

        def changeFilter():
            for lvl, lvl_name in self.levels.items():
                if lvl_name == self.filterBox.currentText():
                    self.table_filter.setLevel(lvl)
                    break

            self.tableView.update()

        self.filterBox.currentIndexChanged.connect(changeFilter)
        self.filterBox.setCurrentText("Warning")

    @QtCore.pyqtSlot(QtCore.QModelIndex, int, int)
    def newLogRecord(self, idx, first, last):
        self.tableView.scrollToBottom()

        record = self.table_filter.sourceModel().log_records[idx.row()]
        if record.levelno >= 30 and self.autoBox.isChecked():
            self.show()


DEFAULT_CSS = """
QRangeSlider * {
    border: 0px;
    padding: 0px;
}
QRangeSlider #Head {
    background: #222;
}
QRangeSlider #Span {
    background: #393;
}
QRangeSlider #Span:active {
    background: #282;
}
QRangeSlider #Tail {
    background: #222;
}
QRangeSlider > QSplitter::handle {
    background: #393;
}
QRangeSlider > QSplitter::handle:vertical {
    height: 4px;
}
QRangeSlider > QSplitter::handle:pressed {
    background: #ca5;
}

"""


def scale(val, src, dst):
    """
    Scale the given value from the scale of src to the scale of dst.
    """
    return int(((val - src[0]) / float(src[1] - src[0])) * (dst[1] - dst[0]) + dst[0])


class Ui_Form(object):
    """default range slider form"""

    def setupUi(self, Form):
        Form.setObjectName("QRangeSlider")
        Form.resize(300, 30)
        Form.setStyleSheet(DEFAULT_CSS)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        # self.gridLayout.setMargin(0)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName("gridLayout")

        self._splitter = QtWidgets.QSplitter(Form)
        self._splitter.setMinimumSize(QtCore.QSize(0, 0))
        self._splitter.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self._splitter.setOrientation(QtCore.Qt.Horizontal)
        self._splitter.setObjectName("splitter")

        self._head = QtWidgets.QGroupBox(self._splitter)
        self._head.setTitle("")
        self._head.setObjectName("Head")
        self._handle = QtWidgets.QGroupBox(self._splitter)
        self._handle.setTitle("")
        self._handle.setObjectName("Span")

        self._tail = QtWidgets.QGroupBox(self._splitter)
        self._tail.setTitle("")
        self._tail.setObjectName("Tail")
        self.gridLayout.addWidget(self._splitter, 0, 0, 1, 1)

        # self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        encoding = QtWidgets.QApplication.UnicodeUTF8
        Form.setWindowTitle(
            QtWidgets.QApplication.translate(
                "QRangeSlider", "QRangeSlider", None, encoding
            )
        )


class Element(QtWidgets.QGroupBox):
    def __init__(self, parent, main):
        super().__init__(parent)
        self.main = main

    def setStyleSheet(self, style):
        """redirect style to parent groupbox"""
        self.parent().setStyleSheet(style)

    def textColor(self):
        """text paint color"""
        return getattr(self, "__textColor", QtGui.QColor(125, 125, 125))

    def format(self, value):
        if self.main._formatter is None:
            return str(value)
        return self.main._formatter(value)

    def setTextColor(self, color):
        """set the text paint color"""
        if type(color) == tuple and len(color) == 3:
            color = QtGui.QColor(color[0], color[1], color[2])
        elif type(color) == int:
            color = QtGui.QColor(color, color, color)
        setattr(self, "__textColor", color)

    def paintEvent(self, event):
        """overrides paint event to handle text"""
        qp = QtGui.QPainter()
        qp.begin(self)
        if self.main.drawValues():
            self.drawText(event, qp)
        qp.end()


class Head(Element):
    """area before the handle"""

    def __init__(self, parent, main):
        super().__init__(parent, main)

    def drawText(self, event, qp):
        qp.setPen(self.textColor())
        qp.setFont(QtGui.QFont("Arial", 10))
        qp.drawText(event.rect(), QtCore.Qt.AlignLeft, self.format(self.main.min()))


class Tail(Element):
    """area after the handle"""

    def __init__(self, parent, main):
        super().__init__(parent, main)

    def drawText(self, event, qp):
        qp.setPen(self.textColor())
        qp.setFont(QtGui.QFont("Arial", 10))
        qp.drawText(event.rect(), QtCore.Qt.AlignRight, self.format(self.main.max()))


class Handle(Element):
    """handle area"""

    def __init__(self, parent, main):
        super().__init__(parent, main)

    def drawText(self, event, qp):
        qp.setPen(self.textColor())
        qp.setFont(QtGui.QFont("Arial", 10))
        qp.drawText(event.rect(), QtCore.Qt.AlignLeft, self.format(self.main.start()))
        qp.drawText(event.rect(), QtCore.Qt.AlignRight, self.format(self.main.end()))

    def mouseMoveEvent(self, event):
        event.accept()
        mx = event.globalX()
        _mx = getattr(self, "__mx", None)

        if not _mx:
            setattr(self, "__mx", mx)
            dx = 0
        else:
            dx = mx - _mx

        setattr(self, "__mx", mx)

        if dx == 0:
            event.ignore()
            return
        elif dx > 0:
            dx = 1
        elif dx < 0:
            dx = -1

        s = self.main.start() + dx
        e = self.main.end() + dx
        if s >= self.main.min() and e <= self.main.max():
            self.main.setRange(s, e)


class QRangeSlider(QtWidgets.QWidget, Ui_Form):
    """
    The QRangeSlider class implements a horizontal range slider widget.

    Inherits QWidget.

    Methods

        * __init__ (self, QWidget parent = None)
        * bool drawValues (self)
        * int end (self)
        * (int, int) getRange (self)
        * int max (self)
        * int min (self)
        * int start (self)
        * setBackgroundStyle (self, QString styleSheet)
        * setDrawValues (self, bool draw)
        * setEnd (self, int end)
        * setStart (self, int start)
        * setRange (self, int start, int end)
        * setSpanStyle (self, QString styleSheet)

    Signals

        * endValueChanged (int)
        * maxValueChanged (int)
        * minValueChanged (int)
        * startValueChanged (int)

    Customizing QRangeSlider

    You can style the range slider as below:
    ::
        QRangeSlider * {
            border: 0px;
            padding: 0px;
        }
        QRangeSlider #Head {
            background: #222;
        }
        QRangeSlider #Span {
            background: #393;
        }
        QRangeSlider #Span:active {
            background: #282;
        }
        QRangeSlider #Tail {
            background: #222;
        }

    Styling the range slider handles follows QSplitter options:
    ::
        QRangeSlider > QSplitter::handle {
            background: #393;
        }
        QRangeSlider > QSplitter::handle:vertical {
            height: 4px;
        }
        QRangeSlider > QSplitter::handle:pressed {
            background: #ca5;
        }
    """

    endValueChanged = QtCore.pyqtSignal(int)
    maxValueChanged = QtCore.pyqtSignal(int)
    minValueChanged = QtCore.pyqtSignal(int)
    startValueChanged = QtCore.pyqtSignal(int)

    # define splitter indices
    _SPLIT_START = 1
    _SPLIT_END = 2

    # signals
    minValueChanged = QtCore.pyqtSignal(int)
    maxValueChanged = QtCore.pyqtSignal(int)
    startValueChanged = QtCore.pyqtSignal(int)
    endValueChanged = QtCore.pyqtSignal(int)

    def __init__(self, parent=None):
        """Create a new QRangeSlider instance.

        :param parent: QWidget parent
        :return: New QRangeSlider instance.

        """
        super().__init__(parent)
        self.setupUi(self)
        self.setMouseTracking(False)

        # self._splitter.setChildrenCollapsible(False)
        self._splitter.splitterMoved.connect(self._handleMoveSplitter)

        # head layout
        self._head_layout = QtWidgets.QHBoxLayout()
        self._head_layout.setSpacing(0)
        # self._head_layout.setMargin(0)
        self._head.setLayout(self._head_layout)
        self.head = Head(self._head, main=self)
        self._head_layout.addWidget(self.head)

        # handle layout
        self._handle_layout = QtWidgets.QHBoxLayout()
        self._handle_layout.setSpacing(0)
        # self._handle_layout.setMargin(0)
        self._handle.setLayout(self._handle_layout)
        self.handle = Handle(self._handle, main=self)
        self.handle.setTextColor((150, 255, 150))
        self._handle_layout.addWidget(self.handle)

        # tail layout
        self._tail_layout = QtWidgets.QHBoxLayout()
        self._tail_layout.setSpacing(0)
        # self._tail_layout.setMargin(0)
        self._tail.setLayout(self._tail_layout)
        self.tail = Tail(self._tail, main=self)
        self._tail_layout.addWidget(self.tail)

        self._formatter = None

        # defaults
        self.setMin(0)
        self.setMax(99)
        self.setStart(0)
        self.setEnd(99)
        self.setDrawValues(True)

    def min(self):
        """:return: minimum value"""
        return getattr(self, "__min", None)

    def max(self):
        """:return: maximum value"""
        return getattr(self, "__max", None)

    def setMin(self, value):
        """sets minimum value"""
        assert type(value) is int
        setattr(self, "__min", value)
        self.minValueChanged.emit(value)

    def setMax(self, value):
        """sets maximum value"""
        assert type(value) is int
        setattr(self, "__max", value)
        self.maxValueChanged.emit(value)

    def start(self):
        """:return: range slider start value"""
        return getattr(self, "__start", None)

    def end(self):
        """:return: range slider end value"""
        return getattr(self, "__end", None)

    def _setStart(self, value):
        """stores the start value only"""
        setattr(self, "__start", value)
        self.startValueChanged.emit(value)

    def setFormatter(self, func):
        self._formatter = func

    def setStart(self, value):
        """sets the range slider start value"""
        assert type(value) is int
        v = self._valueToPos(value)
        self._splitter.splitterMoved.disconnect()
        self._splitter.moveSplitter(v, self._SPLIT_START)
        self._splitter.splitterMoved.connect(self._handleMoveSplitter)
        self._setStart(value)

    def _setEnd(self, value):
        """stores the end value only"""
        setattr(self, "__end", value)
        self.endValueChanged.emit(value)

    def setEnd(self, value):
        """set the range slider end value"""
        assert type(value) is int
        v = self._valueToPos(value)
        self._splitter.splitterMoved.disconnect()
        self._splitter.moveSplitter(v, self._SPLIT_END)
        self._splitter.splitterMoved.connect(self._handleMoveSplitter)
        self._setEnd(value)

    def drawValues(self):
        """:return: True if slider values will be drawn"""
        return getattr(self, "__drawValues", None)

    def setDrawValues(self, draw):
        """sets draw values boolean to draw slider values"""
        assert type(draw) is bool
        setattr(self, "__drawValues", draw)

    def getRange(self):
        """:return: the start and end values as a tuple"""
        return (self.start(), self.end())

    def setRange(self, start, end):
        """set the start and end values"""
        self.setStart(start)
        self.setEnd(end)

    def keyPressEvent(self, event):
        """overrides key press event to move range left and right"""
        key = event.key()
        if key == QtCore.Qt.Key_Left:
            s = self.start() - 1
            e = self.end() - 1
        elif key == QtCore.Qt.Key_Right:
            s = self.start() + 1
            e = self.end() + 1
        else:
            event.ignore()
            return
        event.accept()
        if s >= self.min() and e <= self.max():
            self.setRange(s, e)

    def setBackgroundStyle(self, style):
        """sets background style"""
        self._tail.setStyleSheet(style)
        self._head.setStyleSheet(style)

    def setSpanStyle(self, style):
        """sets range span handle style"""
        self._handle.setStyleSheet(style)

    def _valueToPos(self, value):
        """converts slider value to local pixel x coord"""
        return scale(value, (self.min(), self.max()), (0, self.width()))

    def _posToValue(self, xpos):
        """converts local pixel x coord to slider value"""
        return scale(xpos, (0, self.width()), (self.min(), self.max()))

    def _handleMoveSplitter(self, xpos, index):
        """private method for handling moving splitter handles"""
        hw = self._splitter.handleWidth()

        def _lockWidth(widget):
            width = widget.size().width()
            widget.setMinimumWidth(width)
            widget.setMaximumWidth(width)

        def _unlockWidth(widget):
            widget.setMinimumWidth(0)
            widget.setMaximumWidth(16777215)

        v = self._posToValue(xpos)

        if index == self._SPLIT_START:
            _lockWidth(self._tail)
            if v >= self.end():
                return

            offset = -20
            w = xpos + offset
            self._setStart(v)

        elif index == self._SPLIT_END:
            _lockWidth(self._head)
            if v <= self.start():
                return

            offset = -40
            w = self.width() - xpos + offset
            self._setEnd(v)

        _unlockWidth(self._tail)
        _unlockWidth(self._head)
        _unlockWidth(self._handle)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from hashlib import sha1

import numpy as np
import scipy as sp

try:
    from scipy import fftpack as fft
except ImportError:
    from scipy import fft

import time

from pyrocko import guts
from pyrocko.guts_array import Array

from kite import covariance_ext
from kite.util import property_cached  # noqa
from kite.util import Subject, derampMatrix, squareMatrix, trimMatrix

__all__ = ["Covariance", "CovarianceConfig"]

NOISE_PATCH_MIN_PX = 256 * 256
NOISE_PATCH_MAX_NAN = 0.8

noise_regimes = [
    (1.0 / 2000, np.inf),
    (1.0 / 2000, 1.0 / 500),
    (1.0 / 500, 1.0 / 10),
    (0, np.inf),
]


def modelCovarianceExponential(distance, a, b):
    """Exponential function model to approximate a positive-definite covariance

    We assume the following simple covariance model to describe the empirical
    noise observations:

    .. math::

        cov(d) = c \\cdot e^{\\frac{-d}{b}}

    :param distance: Distance between
    :type distance: float or :class:`numpy.ndarray`
    :param a: Linear model coefficient
    :type a: float
    :param b: Exponential model coefficient
    :type b: float
    :returns: Covariance at ``distance``
    :rtype: :class:`numpy.ndarray`
    """
    return a * np.exp(-distance / b)


def modelCovarianceExponentialCosine(distance, a, b, c, d):
    r"""Exponential function model to approximate a positive-definite covariance

    We assume the following simple covariance model to describe the empirical
    noise observations:

    .. math::

        cov(d) = c \\cdot e^{\\frac{-d}{b}} \\cdot \cos{\\frac{d-c}{d}}

    :param distance: Distance between
    :type distance: float or :class:`numpy.ndarray`
    :param a: Linear model coefficient
    :type a: float
    :param b: Exponential model coefficient
    :type b: float
    :param c: Cosinus distance correction
    :type c: float
    :param c: Cosinus coefficient
    :type c: float
    :returns: Covariance at ``distance``
    :rtype: :class:`numpy.ndarray`
    """
    return a * np.exp(-distance / b) * np.cos((distance - c) / d)


def modelPowerspec(k, beta, D):
    """Exponential linear model to estimate a log-linear power spectrum

    We assume the following log-linear model for a measured power spectrum:

    .. math::

        pow(k) = \\frac{k^\\beta}{D}


    :param k: Wavenumber
    :type k: float or :class:`numpy.ndarray`
    :param a: Exponential model factor
    :type a: float
    :param b: Fractional model factor
    :type b: float
    """
    return (k**beta) / D


class CovarianceConfig(guts.Object):
    noise_coord = Array.T(
        shape=(None,),
        dtype=float,
        serialize_as="list",
        optional=True,
        help="Noise patch coordinates and size,",
    )
    model_coefficients = guts.Tuple.T(
        optional=True,
        help="Covariance model coefficients. Either two (exponential) "
        "or three (exponential and cosine term) coefficients."
        "See also :func:`~kite.covariance.modelCovariance`.",
    )
    model_function = guts.StringChoice.T(
        choices=["exponential", "exponential_cosine"],
        default="exponential",
        help="Covariance approximation function.",
    )
    sampling_method = guts.StringChoice.T(
        choices=["spectral", "spatial"],
        default="spatial",
        help="Method for estimating the covariance and structure function.",
    )
    spatial_bins = guts.Int.T(
        default=75, help="Number of distance bins for spatial covariance sampling."
    )
    spatial_pairs = guts.Int.T(
        default=200000, help="Number of random pairs for spatial covariance sampling."
    )
    variance = guts.Float.T(optional=True, help="Variance of the model.")
    adaptive_subsampling = guts.Bool.T(
        default=True, help="Adaptive subsampling flag for full covariance calculation."
    )
    covariance_matrix = Array.T(
        dtype=float,
        optional=True,
        serialize_as="base64",
        help="Cached covariance matrix, "
        "see :attr:`~kite.Covariance.covariance_matrix`.",
    )

    def __init__(self, *args, **kwargs):
        if len(kwargs) != 0:
            if "a" in kwargs and "b" in kwargs:
                kwargs["model_coefficients"] = (kwargs.pop("a"), kwargs.pop("b"))
        guts.Object.__init__(self, *args, **kwargs)


class Covariance(object):
    """Construct the variance-covariance matrix of quadtree subsampled data.

    Variance and covariance estimates are used to construct the weighting
    matrix to be used later in an optimization.

    Two different methods exist to propagate full-resolution data variances
    and covariances of :class:`kite.Scene.displacement` to the
    covariance matrix of the subsampled dataset:

    1. The distance between :py:class:`kite.quadtree.QuadNode`
       leaf focal points, :py:class:`kite.covariance.Covariance.matrix_focal`
       defines the approximate covariance of the quadtree leaf pair.
    2. The _accurate_ propagation of covariances by taking the mean of
       every node pair pixel covariances. This process is computational
       very expensive and can take a few minutes.
       :py:class:`kite.covariance.Covariance.matrix_focal`

    :param quadtree: Quadtree to work on
    :type quadtree: :class:`~kite.Quadtree`
    :param config: Config object
    :type config: :class:`~kite.covariance.CovarianceConfig`
    """

    def __init__(self, scene, config=CovarianceConfig()):
        self.evChanged = Subject()
        self.evConfigChanged = Subject()

        self.frame = scene.frame
        self.quadtree = scene.quadtree
        self.scene = scene
        self.nthreads = 0
        self._noise_data = None
        self._powerspec1d_cached = None
        self._powerspec2d_cached = None
        self._powerspec3d_cached = None
        self._noise_data_grid = None
        self._initialized = False
        self._log = scene._log.getChild("Covariance")

        self.setConfig(config)
        self.quadtree.evChanged.subscribe(self._clear)
        self.scene.evConfigChanged.subscribe(self.setConfig)

    def __call__(self, *args, **kwargs):
        return self.getLeafCovariance(*args, **kwargs)

    def setConfig(self, config=None):
        """Sets and updated the config of the instance

        :param config: New config instance, defaults to configuration provided
                       by parent :class:`~kite.Scene`
        :type config: :class:`~kite.covariance.CovarianceConfig`, optional
        """
        if config is None:
            config = self.scene.config.covariance

        if self.scene.config.old_import:
            self._log.warning("Old format - resetting noise patch coordinates")
            config.covariance_matrix = None
            config.noise_coord = None

        self.config = config
        if config.noise_coord is None and (
            config.model_coefficients is not None or config.variance is not None
        ):
            self.noise_data  # init data array
            self.config.model_coefficients = config.model_coefficients
            self.config.variance = config.variance

        self._clear(config=False)
        self.evConfigChanged.notify()

    def _clear(self, config=True, spectrum=True):
        if config:
            self.config.model_coefficients = None
            self.config.variance = None
            self.config.covariance_matrix = None

        if spectrum:
            self.structure_spectral = None
            self._powerspec1d_cached = None
            self._powerspec2d_cached = None

        self._noise_data_grid = None
        self.covariance_matrix = None
        self.covariance_matrix_focal = None
        self.covariance_spectral = None
        self.covariance_spatial = None
        self.structure_spatial = None
        self.weight_matrix = None
        self.weight_matrix_focal = None
        self._initialized = False
        self.evChanged.notify()

    @property
    def finished_combinations(self):
        return covariance_ext.get_finished_combinations()

    @property
    def noise_coord(self):
        """Coordinates of the noise patch in local coordinates.

        :setter: Set the noise coordinates
        :getter: Get the noise coordinates
        :type: :class:`numpy.ndarray`, ``[llE, llN, sizeE, sizeN]``
        """
        if self.config.noise_coord is None:
            self.noise_data
        return self.config.noise_coord

    @noise_coord.setter
    def noise_coord(self, values):
        self.config.noise_coord = np.array(values)

    @property
    def noise_patch_size_km2(self):
        """
        :getter: Noise patch size in :math:`km^2`.
        :type: float
        """
        if self.noise_coord is None:
            return 0.0
        size = (self.noise_coord[2] * self.noise_coord[3]) * 1e-6
        if self.noise_data.size < self.NOISE_PATCH_MIN_PX:
            self._log.warning("Defined noise patch is instably small")
        return size

    @property
    def noise_data(self, data):
        """Noise data we process to estimate the covariance

        :setter: Set the noise patch to analyze the covariance.
        :getter: If the noise data has not been set manually, we grab data
                 through :func:`~kite.Covariance.selectNoiseNode`.
        :type: :class:`numpy.ndarray`
        """
        return self._noise_data

    @noise_data.getter
    def noise_data(self):
        if self._noise_data is not None:
            return self._noise_data
        elif self.config.noise_coord is not None:
            self._log.debug("Selecting noise_data from config...")
            llE, llN = self.scene.frame.mapENMatrix(*self.config.noise_coord[:2])
            sE, sN = self.scene.frame.mapENMatrix(*self.config.noise_coord[2:])
            slice_E = slice(llE, llE + sE)
            slice_N = slice(llN, llN + sN)

            covariance_matrix = self.config.covariance_matrix
            self.noise_data = self.scene.displacement[slice_N, slice_E]
            self.config.covariance_matrix = covariance_matrix
        else:
            self._log.debug("Selecting noise_data from Quadtree...")
            node = self.selectNoiseNode()
            self.noise_data = node.displacement
            self.noise_coord = [node.llE, node.llN, node.sizeE, node.sizeN]

        return self._noise_data

    @noise_data.setter
    def noise_data(self, data):
        data = data.copy()
        data = derampMatrix(trimMatrix(data))
        data[np.isnan(data)] = 0.0
        self._noise_data = data
        self._clear()

    @property
    def noise_data_gridE(self):
        return self._get_noise_data_grid()[0]

    @property
    def noise_data_gridN(self):
        return self._get_noise_data_grid()[1]

    def _get_noise_data_grid(self):
        if self._noise_data_grid is None:
            scene = self.scene

            llE, llN = scene.frame.mapENMatrix(*self.noise_coord[:2])
            sE, sN = scene.frame.mapENMatrix(*self.noise_coord[2:])
            slice_E = slice(llE, llE + sE + 1)
            slice_N = slice(llN, llN + sN + 1)

            gridE = scene.frame.gridEmeter[slice_N, slice_E]
            gridN = scene.frame.gridNmeter[slice_N, slice_E]

            gridE = trimMatrix(self.noise_data, data=gridE)
            gridN = trimMatrix(self.noise_data, data=gridN)

            self._noise_data_grid = (gridE, gridN)

        return self._noise_data_grid

    def selectNoiseNode(self):
        """Choose noise node from quadtree
        the biggest :class:`~kite.quadtree.QuadNode` from
        :class:`~kite.Quadtree`.

        :returns: A quadnode with the least signal.
        :rtype: :class:`~kite.quadtree.QuadNode`
        """
        t0 = time.time()

        node_selection = [
            n
            for n in self.quadtree.nodes
            if n.npixel > NOISE_PATCH_MIN_PX and n.nan_fraction < NOISE_PATCH_MAX_NAN
        ]
        if not node_selection:
            node_selection = self.quadtree.leaves

        stdmax = max([n.std for n in node_selection])
        lmax = max([n.std for n in node_selection])

        def costFunction(n):
            nl = np.log2(n.length) / np.log2(lmax)
            ns = n.std / stdmax
            return nl * (1.0 - ns) * (1.0 - n.nan_fraction)

        fitness = np.array([costFunction(n) for n in node_selection])

        self._log.debug(
            "Fetched noise from Quadtree.nodes [%0.4f s]" % (time.time() - t0)
        )
        node = node_selection[np.argmin(fitness)]
        return node

    def _mapLeaves(self, nx, ny):
        """Helper function returning appropriate
            :class:`~kite.quadtree.QuadNode` and for maintaining
            the internal mapping with the matrices.

        :param nx: matrix x position
        :type nx: int
        :param ny: matrix y position
        :type ny: int
        :returns: tuple of :class:`~kite.quadtree.QuadNode` s for ``nx``
            and ``ny``
        :rtype: tuple
        """
        leaf1 = self.quadtree.leaves[nx]
        leaf2 = self.quadtree.leaves[ny]

        self._leaf_mapping[leaf1.id] = nx
        self._leaf_mapping[leaf2.id] = ny

        return leaf1, leaf2

    def isFullCovarianceCalculated(self):
        if self.config.covariance_matrix is None:
            return False
        return True

    @property_cached
    def covariance_matrix(self):
        """Covariance matrix calculated from mean of all pixel pairs
            inside the node pairs (full and accurate propagation).

        :type: :class:`numpy.ndarray`,
            size (:class:`~kite.Quadtree.nleaves` x
            :class:`~kite.Quadtree.nleaves`)
        """
        if not isinstance(self.config.covariance_matrix, np.ndarray):
            self.config.covariance_matrix = self._calcCovarianceMatrix(method="full")
            self.evChanged.notify()
        elif self.config.covariance_matrix.ndim == 1:
            try:
                nl = self.quadtree.nleaves
                self.config.covariance_matrix = self.config.covariance_matrix.reshape(
                    nl, nl
                )
            except ValueError:
                self.config.covariance_matrix = None
                return self.covariance_matrix
        return self.config.covariance_matrix

    @property_cached
    def covariance_matrix_focal(self):
        """Approximate Covariance matrix from quadtree leaf pair
            distance only. Fast, use for intermediate steps only and
            finally use approach :attr:`~kite.Covariance.covariance_matrix`.

        :type: :class:`numpy.ndarray`,
            size (:class:`~kite.Quadtree.nleaves` x
            :class:`~kite.Quadtree.nleaves`)
        """
        return self._calcCovarianceMatrix(method="focal")

    @property_cached
    def weight_matrix(self):
        """Weight matrix from full covariance :math:`cov^{-1}`.

        :type: :class:`numpy.ndarray`,
            size (:class:`~kite.Quadtree.nleaves` x
            :class:`~kite.Quadtree.nleaves`)
        """
        return np.linalg.inv(self.covariance_matrix)

    @property_cached
    def weight_matrix_L2(self):
        """Weight matrix from full covariance :math:`\\sqrt{cov^{-1}}`.

        :type: :class:`numpy.ndarray`,
            size (:class:`~kite.Quadtree.nleaves` x
            :class:`~kite.Quadtree.nleaves`)
        """
        incov = np.linalg.inv(self.covariance_matrix)
        return sp.linalg.sqrtm(incov)

    @property_cached
    def weight_matrix_focal(self):
        """Approximated weight matrix from fast focal method
            :math:`cov_{focal}^{-1}`.

        :type: :class:`numpy.ndarray`,
            size (:class:`~kite.Quadtree.nleaves` x
            :class:`~kite.Quadtree.nleaves`)
        """
        try:
            return np.linalg.inv(self.covariance_matrix_focal)
        except np.linalg.LinAlgError as e:
            self._log.exception(e)
            return np.eye(self.covariance_matrix_focal.shape[0])

    @property_cached
    def weight_vector(self):
        """Weight vector from full covariance :math:`cov^{-1}`.
        :type: :class:`numpy.ndarray`,
            size (:class:`~kite.Quadtree.nleaves`)
        """
        return np.sum(self.weight_matrix, axis=1)

    @property_cached
    def weight_vector_focal(self):
        """Weight vector from fast focal method
            :math:`\\sqrt{cov_{focal}^{-1}}`.
        :type: :class:`numpy.ndarray`,
            size (:class:`~kite.Quadtree.nleaves`)
        """
        return np.sum(self.weight_matrix_focal, axis=1)

    def _calcCovarianceMatrix(self, method="focal", nthreads=None):
        """Constructs the covariance matrix.

        :param method: Either ``focal`` point distances are used - this is
            quick but only an approximation.
            Or ``full``, where the full quadtree pixel distances matrices are
            calculated , defaults to ``focal``
        :type method: str, optional
        :returns: Covariance matrix
        :rtype: thon:numpy.ndarray
        """
        self._initialized = True
        nthreads = nthreads or self.nthreads

        nl = len(self.quadtree.leaves)
        self._leaf_mapping = {}

        t0 = time.time()

        if method == "focal":
            model = self.getModelFunction()

            coords = self.quadtree.leaf_focal_points_meter
            dist_matrix = np.sqrt(
                (coords[:, 0] - coords[:, 0, np.newaxis]) ** 2
                + (coords[:, 1] - coords[:, 1, np.newaxis]) ** 2
            )
            cov_matrix = model(dist_matrix, *self.covariance_model)

            # adding variance
            if self.variance < cov_matrix.max():
                variance = cov_matrix.max()
            else:
                variance = self.variance
            if self.quadtree.leaf_mean_px_var is not None:
                self._log.debug("Adding variance from scene.displacement_px_var")
                variance += self.quadtree.leaf_mean_px_var
            np.fill_diagonal(cov_matrix, variance)

            for nx, ny in np.nditer(np.triu_indices_from(dist_matrix)):
                self._mapLeaves(nx, ny)

        elif method == "full":
            leaf_map = np.empty((len(self.quadtree.leaves), 4), dtype=np.uint32)
            for nl, leaf in enumerate(self.quadtree.leaves):
                leaf, _ = self._mapLeaves(nl, nl)
                leaf_map[nl, 0], leaf_map[nl, 1] = (
                    leaf._slice_rows.start,
                    leaf._slice_rows.stop,
                )
                leaf_map[nl, 2], leaf_map[nl, 3] = (
                    leaf._slice_cols.start,
                    leaf._slice_cols.stop,
                )

            nleaves = self.quadtree.nleaves
            cov_matrix = covariance_ext.covariance_matrix(
                self.scene.frame.gridEmeter.filled(),
                self.scene.frame.gridNmeter.filled(),
                leaf_map,
                self.covariance_model,
                self.variance,
                nthreads,
                self.config.adaptive_subsampling,
            ).reshape(nleaves, nleaves)

            if self.quadtree.leaf_mean_px_var is not None:
                self._log.debug("Adding variance from scene.displacement_px_var")
                cov_matrix[
                    np.diag_indices_from(cov_matrix)
                ] += self.quadtree.leaf_mean_px_var

        else:
            raise TypeError("Covariance calculation %s method not defined!" % method)

        self._log.debug(
            "Created covariance matrix - %s mode [%0.4f s]" % (method, time.time() - t0)
        )
        return cov_matrix

    def isMatrixPosDefinite(self, full=False):
        self._log.debug("Checking whether matrix is positive-definite")
        if full:
            matrix = self.covariance_matrix
        else:
            matrix = self.covariance_matrix_focal

        try:
            chol_decomp = np.linalg.cholesky(matrix)
        except np.linalg.linalg.LinAlgError:
            pos_def = False
        else:
            pos_def = ~np.all(np.iscomplex(chol_decomp))
        finally:
            if not pos_def:
                self._log.warning("Covariance matrix is not positive definite!")
            return pos_def

    @staticmethod
    def _leafFocalDistance(leaf1, leaf2):
        return np.sqrt(
            (leaf1.focal_point[0] - leaf2.focal_point[0]) ** 2
            + (leaf1.focal_point[1] - leaf2.focal_point[1]) ** 2
        )

    def _leafMapping(self, leaf1, leaf2):
        if not isinstance(leaf1, str):
            leaf1 = leaf1.id
        if not isinstance(leaf2, str):
            leaf2 = leaf2.id
        if not self._initialized:
            self.covariance_matrix_focal
        try:
            return self._leaf_mapping[leaf1], self._leaf_mapping[leaf2]
        except KeyError as e:
            raise KeyError("Unknown quadtree leaf with id %s" % e)

    def getLeafCovariance(self, leaf1, leaf2):
        """Get the covariance between ``leaf1`` and ``leaf2`` from
            distances.

        :param leaf1: Leaf one
        :type leaf1: str of `leaf.id` or :class:`~kite.quadtree.QuadNode`
        :param leaf2: Leaf two
        :type leaf2: str of `leaf.id` or :class:`~kite.quadtree.QuadNode`
        :returns: Covariance between ``leaf1`` and ``leaf2``
        :rtype: float
        """
        return self.covariance_matrix[self._leafMapping(leaf1, leaf2)]

    def getLeafWeight(self, leaf, model="focal"):
        """Get the total weight of ``leaf``, which is the summation of
            all single pair weights of :attr:`kite.Covariance.weight_matrix`.

        .. math ::

            w_{x} = \\sum_i W_{x,i}

        :param model: ``Focal`` or ``full``, default ``focal``
        :type model: str
        :param leaf: A leaf from :class:`~kite.Quadtree`
        :type leaf: :class:`~kite.quadtree.QuadNode`

        :returns: Weight of the leaf
        :rtype: float
        """
        (nl, _) = self._leafMapping(leaf, leaf)
        weight_mat = self.weight_matrix_focal
        return np.mean(weight_mat, axis=0)[nl]

    def syntheticNoise(
        self, shape=(1024, 1024), dEdN=None, anisotropic=False, rstate=None
    ):
        """Create random synthetic noise from data noise power spectrum.

        This function uses the power spectrum of the data noise
        (:attr:`noise_data`) (:func:`powerspecNoise`) to create synthetic
        noise, e.g. to use it for data perturbation in optinmizations.
        The default sampling distances are taken from
        :attr:`kite.scene.Frame.dE` and :attr:`kite.scene.Frame.dN`. They can
        be overwritten.

        :param shape: shape of the desired noise patch.
            Pixels in northing and easting (`nE`, `nN`),
            defaults to `(1024, 1024)`.
        :type shape: tuple, optional
        :param dEdN: The sampling distance in east and north [m], defaults to
            (:attr:`kite.scene.Frame.dEmeter`,
             :attr:`kite.scene.Frame.dNmeter`).
        :type dEdN: tuple, floats
        :returns: synthetic noise patch
        :rtype: :class:`numpy.ndarray`
        """
        if (shape[0] + shape[1]) % 2 != 0:
            # self._log.warning('Patch dimensions must be even, '
            #                   'ceiling dimensions!')
            pass
        nE = shape[1] + (shape[1] % 2)
        nN = shape[0] + (shape[0] % 2)

        if rstate is None:
            rstate = np.random.RandomState()

        rfield = rstate.rand(nN, nE)
        spec = np.fft.fft2(rfield)

        if not dEdN:
            dE, dN = (self.scene.frame.dEmeter, self.scene.frame.dNmeter)
        kE = np.fft.fftfreq(nE, dE)
        kN = np.fft.fftfreq(nN, dN)
        k_rad = np.sqrt(kN[:, np.newaxis] ** 2 + kE[np.newaxis, :] ** 2)

        amp = np.zeros_like(k_rad)

        if not anisotropic:
            noise_pspec, k, _, _, _, _ = self.powerspecNoise2D()
            k_bin = np.insert(k + k[0] / 2, 0, 0)

            for i in range(k.size):
                k_min = k_bin[i]
                k_max = k_bin[i + 1]
                r = np.logical_and(k_rad > k_min, k_rad <= k_max)
                if i == (k.size - 1):
                    r = k_rad > k_min
                if not np.any(r):
                    continue
                amp[r] = noise_pspec[i]
            amp[k_rad == 0.0] = self.variance
            amp[k_rad > k.max()] = noise_pspec[np.argmax(k)]
            amp = np.sqrt(amp * self.noise_data.size * np.pi * 4)

        elif anisotropic:
            interp_pspec, _, _, _, skE, skN = self.powerspecNoise3D()
            kE = np.fft.fftshift(kE)
            kN = np.fft.fftshift(kN)
            make = np.logical_and(kE >= skE.min(), kE <= skE.max())
            mkN = np.logical_and(kN >= skN.min(), kN <= skN.max())
            mkRad = np.where(  # noqa
                k_rad < np.sqrt(kN[mkN].max() ** 2 + kE[make].max() ** 2)
            )
            res = interp_pspec(kN[mkN, np.newaxis], kE[np.newaxis, make], grid=True)
            amp = res
            amp = np.fft.fftshift(amp)

        spec *= amp
        noise = np.abs(np.fft.ifft2(spec))
        noise -= np.mean(noise)

        # remove shape % 2 padding
        return noise[: shape[0], : shape[1]]

    def getQuadtreeNoise(self, rstate=None, gather=np.nanmedian):
        """Create noise for a :class:`~kite.quadtree.Quadtree`

        Use :meth:`~kite.covariance.Covariance.getSyntheticNoise` to create
        data-driven noise on each quadtree leaf, summarized by

        :param gather: Function gathering leaf's noise realisation,
                       defaults to np.median.
        :type normalisation: callable, optional
        :returns: Array of noise level at each quadtree leaf.
        :rtype: :class:`numpy.ndarray`
        """
        qt = self.quadtree

        syn_noise = self.syntheticNoise(
            shape=self.scene.displacement.shape, rstate=rstate
        )
        syn_noise[self.scene.displacement_mask] = np.nan
        noise_quadtree_arr = np.full(qt.nleaves, np.nan)

        for il, lv in enumerate(qt.leaves):
            noise_quadtree_arr[il] = gather(syn_noise[lv._slice_rows, lv._slice_cols])
        return noise_quadtree_arr

    def powerspecNoise1D(self, data=None, ndeg=512, nk=512):
        if self._powerspec1d_cached is None:
            self._powerspec1d_cached = self._powerspecNoise(
                data, norm="1d", ndeg=ndeg, nk=nk
            )
        return self._powerspec1d_cached

    def powerspecNoise2D(self, data=None, ndeg=512, nk=512):
        if self._powerspec2d_cached is None:
            self._powerspec2d_cached = self._powerspecNoise(
                data, norm="2d", ndeg=ndeg, nk=nk
            )
        return self._powerspec2d_cached

    def powerspecNoise3D(self, data=None):
        if self._powerspec3d_cached is None:
            self._powerspec3d_cached = self._powerspecNoise(data, norm="3d")
        return self._powerspec3d_cached

    def _powerspecNoise(self, data=None, norm="1d", ndeg=512, nk=512):
        """Get the noise power spectrum from
            :attr:`kite.Covariance.noise_data`.

        :param data: Overwrite Covariance.noise_data, defaults to `None`
        :type data: :class:`numpy.ndarray`, optional
        :returns: `(power_spec, k, f_spectrum, kN, kE)`
        :rtype: tuple
        """
        if data is None:
            noise = self.noise_data
        else:
            noise = data.copy()
        if norm not in ("1d", "2d", "3d"):
            raise AttributeError("norm must be 1d, 2d or 3d")

        # noise = squareMatrix(noise)
        shift = np.fft.fftshift

        spectrum = shift(np.fft.fft2(noise, axes=(0, 1), norm=None))
        power_spec = (np.abs(spectrum) / spectrum.size) ** 2

        kE = shift(np.fft.fftfreq(power_spec.shape[1], d=self.quadtree.frame.dEmeter))
        kN = shift(np.fft.fftfreq(power_spec.shape[0], d=self.quadtree.frame.dNmeter))
        k_rad = np.sqrt(kN[:, np.newaxis] ** 2 + kE[np.newaxis, :] ** 2)
        power_spec[k_rad == 0.0] = 0.0

        power_interp = sp.interpolate.RectBivariateSpline(kN, kE, power_spec)

        # def power1d(k):
        #     theta = np.linspace(-np.pi, np.pi, ndeg, False)
        #     power = np.empty_like(k)
        #     for i in range(k.size):
        #         kE = np.cos(theta) * k[i]
        #         kN = np.sin(theta) * k[i]
        #         power[i] = np.median(power_interp.ev(kN, kE)) * k[i]\
        #             * np.pi * 4
        #     return power

        def power1d(k):
            theta = np.linspace(-np.pi, np.pi, ndeg, False)
            power = np.empty_like(k)

            cos_theta = np.cos(theta)
            sin_theta = np.sin(theta)
            for i in range(k.size):
                kE = cos_theta * k[i]
                kN = sin_theta * k[i]
                power[i] = np.mean(power_interp.ev(kN, kE))

            power *= 2 * np.pi
            return power

        def power2d(k):
            """Mean 2D Power works!"""
            theta = np.linspace(-np.pi, np.pi, ndeg, False)
            power = np.empty_like(k)

            cos_theta = np.cos(theta)
            sin_theta = np.sin(theta)
            for i in range(k.size):
                kE = sin_theta * k[i]
                kN = cos_theta * k[i]
                power[i] = np.median(power_interp.ev(kN, kE))
                # Median is more stable than the mean here

            return power

        def power3d(k):
            return power_interp

        power = power1d
        if norm == "2d":
            power = power2d
        elif norm == "3d":
            power = power3d

        k_rad = np.sqrt(kN[:, np.newaxis] ** 2 + kE[np.newaxis, :] ** 2)
        k = np.linspace(k_rad[k_rad > 0].min(), k_rad.max(), nk)
        dk = 1.0 / (k[1] - k[0]) / (2 * nk)
        return power(k), k, dk, spectrum, kE, kN

    def _powerCosineTransform(self, p_spec):
        """Calculating the cosine transform of the power spectrum.

        The cosine transform of the power spectrum is an estimate
        of the data covariance (see Hanssen, 2001)."""
        cos = fft.idct(p_spec, type=3)
        return cos

    def setSamplingMethod(self, method):
        """Set the sampling method"""
        assert method in CovarianceConfig.sampling_method.choices

        self.config.sampling_method = method
        self._clear(config=True, spectrum=False)
        self.evChanged.notify()
        self._log.debug("Changed sampling method to %s" % method)

    def setSpatialBins(self, nbins):
        """Set number of spatial bins"""
        self.config.spatial_bins = nbins
        self._clear(config=True, spectrum=False)
        self.evChanged.notify()
        self._log.debug("Changed spatial distance bins to %s" % nbins)

    def setSpatialPairs(self, npairs):
        """Set number of random spatial pairs"""
        self.config.spatial_pairs = npairs
        self._clear(config=True, spectrum=False)
        self.evChanged.notify()
        self._log.debug("Changed random pairs to %s" % npairs)

    def setModelFunction(self, model):
        assert model in CovarianceConfig.model_function.choices
        self.config.model_function = model
        self._clear(config=True, spectrum=True)
        self.evChanged.notify()
        self._log.debug("Changed model function to %s" % model)

    def getModelFunction(self):
        if self.config.model_function == "exponential":
            return modelCovarianceExponential
        if self.config.model_function == "exponential_cosine":
            return modelCovarianceExponentialCosine

    @property_cached
    def covariance_spectral(self):
        """Covariance function estimated directly from the power spectrum of
            displacement noise patch using the cosine transform.

        :type: tuple, :class:`numpy.ndarray` (covariance, distance)"""
        power_spec, k, dk, _, _, _ = self.powerspecNoise1D()
        # power_spec -= self.variance

        d = np.arange(1, power_spec.size + 1) * dk
        cov = self._powerCosineTransform(power_spec)

        return cov, d

    @property_cached
    def covariance_spatial(self):
        self._log.debug("Estimating covariance (spatial)...")

        nbins = self.config.spatial_bins
        npairs = self.config.spatial_pairs
        noise_data = self.noise_data.ravel()
        noise_data -= noise_data.mean()

        grdE = self.noise_data_gridE
        grdN = self.noise_data_gridN

        max_distance = min(abs(grdE.min() - grdE.max()), abs(grdN.min() - grdN.max()))
        dist_bins = np.linspace(0, max_distance, nbins + 1)

        grdE = grdE.ravel()
        grdN = grdN.ravel()

        # Select random coordinates
        rstate = np.random.RandomState(noise_data.size)
        rand_idx = rstate.randint(0, noise_data.size, (2, npairs))
        idx0 = rand_idx[0, :]
        idx1 = rand_idx[1, :]

        distances = np.sqrt(
            (grdN[idx0] - grdN[idx1]) ** 2 + (grdE[idx0] - grdE[idx1]) ** 2
        )

        cov_all = noise_data[idx0] * noise_data[idx1]
        vario_all = (noise_data[idx0] - noise_data[idx1]) ** 2

        bins = np.digitize(distances, dist_bins, right=True)
        bin_distances = dist_bins[1:] - dist_bins[1] / 2

        covariance = np.full_like(bin_distances, fill_value=np.nan)
        variance = np.full_like(bin_distances, fill_value=np.nan)

        for ib in range(nbins):
            selection = bins == ib
            if selection.sum() != 0:
                covariance[ib] = np.nanmean(cov_all[selection])
                variance[ib] = np.nanmean(vario_all[selection]) / 2

        self._structure_spatial = (
            variance[~np.isnan(variance)],
            bin_distances[~np.isnan(variance)],
        )
        covariance[0] = np.nan
        return (
            covariance[~np.isnan(covariance)],
            bin_distances[~np.isnan(covariance)],
        )

    def getCovariance(self):
        """Calculate the covariance function

        :return: The covariance and distance
        :rtype: tuple
        """
        if self.config.sampling_method == "spatial":
            return self.covariance_spatial
        elif self.config.sampling_method == "spectral":
            return self.covariance_spectral

    @property
    def covariance_model(self, regime=0):
        """Covariance model parameters for
            :func:`~kite.covariance.modelCovariance` retrieved
            from :attr:`~kite.Covariance.getCovarianceFunction`.

        .. note:: using this function implies several model
            fits: (1) fit of the spectrum and (2) the cosine transform.
            Not sure about the consequences, if this is useful and/or
            meaningful.

        :getter: Get the parameters.
        :type: tuple, ``a`` and ``b``
        """
        if self.config.model_coefficients is None:
            covariance, distance = self.getCovariance()
            model = self.getModelFunction()

            if self.config.model_function == "exponential":
                coeff = (np.mean(covariance), np.mean(distance))

            elif self.config.model_function == "exponential_cosine":
                coeff = (
                    np.mean(covariance),
                    np.mean(distance),
                    np.mean(distance) * -0.1,
                    0.1,
                )

                func = self.getModelFunction()

                # Testing penalty function
                def model(*args):
                    distance, a, b, c, d = args
                    res = func(*args)

                    penalty = 0.0
                    if distance[-1] / b > (distance[-1] + c) / d:
                        penalty = (b - d) * coeff[0]
                        self._log.warning("Penalty %f" % penalty)

                    return res + penalty

                # Overwrite with pure model function
                model = self.getModelFunction()  # noqa

            try:
                coeff, _ = sp.optimize.curve_fit(model, distance, covariance, p0=coeff)
            except (RuntimeError, TypeError) as e:
                self._log.exception(e)
                self._log.warning(
                    "Could not fit the %s covariance model", self.config.model_function
                )
            finally:
                self.config.model_coefficients = tuple(map(float, coeff))

        return self.config.model_coefficients

    @property
    def covariance_model_rms(self):
        """
        :getter: RMS missfit between :class:`~kite.Covariance.covariance_model`
            and :class:`~kite.Covariance.getCovarianceFunction`
        :type: float
        """
        cov, d = self.getCovariance()
        model = self.getModelFunction()
        cov_mod = model(d, *self.covariance_model)

        return np.sqrt(np.mean((cov - cov_mod) ** 2))

    @property_cached
    def structure_spatial(self):
        self.covariance_spatial
        return self._structure_spatial

    @property_cached
    def structure_spectral(self):
        """Structure function derived from ``noise_patch``
            :type: tuple, :class:`numpy.ndarray` (structure_spectral, distance)

        Adapted from
        http://clouds.eos.ubc.ca/~phil/courses/atsc500/docs/strfun.pdf
        """
        power_spec, k, dk, _, _, _ = self.powerspecNoise1D()
        d = np.arange(1, power_spec.size + 1) * dk

        def structure_spectral(power_spec, d, k):
            struc_func = np.zeros_like(k)
            for i, d in enumerate(d):
                for ik, tk in enumerate(k):
                    # struc_func[i] += (1. - np.cos(tk*d))*power_spec[ik]
                    struc_func[i] += (1.0 - sp.special.j0(tk * d)) * power_spec[ik]
            struc_func *= 2.0 / 1
            return struc_func

        struc_func = structure_spectral(power_spec, d, k)
        return struc_func, d

    def getStructure(self, method=None):
        """Get the structure function

        :param method: Either `spatial` or `spectral`, if `None`
            the method is taken from config
        :type method: str (optional)

        :return: (variance, distance)
        :rtype: tuple
        """
        if method is None:
            method = self.config.sampling_method
        if method == "spatial":
            return self.structure_spatial
        elif method == "spectral":
            return self.structure_spectral

    @property
    def variance(self):
        """Variance of data noise estimated from the
            high-frequency end of power spectrum.

        :setter: Set the variance manually
        :getter: Retrieve the variance
        :type: float
        """
        return self.config.variance

    @variance.setter
    def variance(self, value):
        self.config.variance = float(value)
        # self._clear(config=False, spectrum=False, spatial=False)
        self.evChanged.notify()

    @variance.getter
    def variance(self):

        if self.config.variance is None and self.config.sampling_method == "spatial":
            structure_spatial, dist = self.structure_spatial

            last_20p = -int(structure_spatial.size * 0.2)
            self.config.variance = float(np.mean(structure_spatial[(last_20p):]))

        elif self.config.variance is None and self.config.sampling_method == "spectral":
            power_spec, k, dk, spectrum, _, _ = self.powerspecNoise1D()
            cov, _ = self.covariance_spectral
            ma = self.covariance_model[0]
            # print(cov[1])
            ps = power_spec * spectrum.size
            # print(spectrum.size)
            # print(np.mean(ps[-int(ps.size/9.):-1]))
            var = np.median(ps[-int(ps.size / 9.0) :]) + ma
            self.config.variance = float(var)

        return self.config.variance

    def export_weight_matrix(self, filename):
        """Export the full :attr:`~kite.Covariance.weight_matrix` to an ASCII
            file. The data can be loaded through :func:`numpy.loadtxt`.

        :param filename: path to export to
        :type filename: str
        """
        self._log.debug("Exporting Covariance.weight_matrix to %s" % filename)
        header = (
            "Exported kite.Covariance.weight_matrix, "
            "for more information visit https://pyrocko.org\n"
            "\nThe matrix is symmetric and ordered by QuadNode.id:\n"
        )
        header += ", ".join([l.id for l in self.quadtree.leaves])
        np.savetxt(filename, self.weight_matrix, header=header)

    def get_state_hash(self):
        sha = sha1()
        sha.update(str(self.config).encode())
        return sha.digest().hex()

    @property_cached
    def plot(self):
        """Simple overview plot to summarize the covariance estimations."""
        from kite.plot2d import CovariancePlot

        return CovariancePlot(self)

    @property_cached
    def plot_syntheticNoise(self):
        """Simple overview plot to summarize the covariance estimations."""
        from kite.plot2d import SyntheticNoisePlot

        return SyntheticNoisePlot(self)



#!/usr/bin/python2
from collections import OrderedDict

import numpy as np
import pyqtgraph as pg
import pyqtgraph.parametertree.parameterTypes as pTypes
from PyQt5 import QtCore, QtWidgets

from kite.qt_utils import loadUi

from .base import KiteParameterGroup, KitePlot, KiteView, get_resource

__all__ = ["KiteScene"]

km = 1e3
pen_roi = pg.mkPen((255, 23, 68), width=2)
pen_roi_highlight = pg.mkPen((115, 210, 22), width=2, style=QtCore.Qt.DashLine)


class PolygonMaskROI(pg.PolyLineROI):
    def _makePen(self):
        # Generate the pen color for this ROI based on its current state.
        if self.mouseHovering:
            return pen_roi_highlight
        else:
            return self.pen


class KiteScene(KiteView):
    title = "Scene"

    def __init__(self, spool):
        model = spool.model
        self.model = model
        self.spool = spool

        scene_plot = KiteScenePlot(model)
        self.scene_plot = scene_plot
        self.main_widget = scene_plot
        self.tools = {
            # 'Components': KiteToolComponents(self.main_widget),
            # 'Displacement Transect': KiteToolTransect(self.main_widget),
        }

        self.param_scene = ParamScene(model, scene_plot)
        self.param_frame = ParamSceneFrame(model, expanded=False)
        self.param_meta = ParamSceneMeta(model, expanded=False)

        self.param_scene.addChild(self.param_frame)
        self.param_scene.addChild(self.param_meta)

        self.deramp_ctrl = DerampParams(model)

        self.parameters = [self.param_scene, self.deramp_ctrl]

        self.dialogTransect = KiteToolTransect(scene_plot, spool)

        spool.actionTransect.triggered.connect(self.dialogTransect.show)

        spool.actionAddPolygonMask.triggered.connect(scene_plot.newMaskPolygon)
        spool.actionTogglePolygonMask.setChecked(
            model.getScene().polygon_mask.is_enabled()
        )
        spool.actionTogglePolygonMask.toggled.connect(self.togglePolygonMask)

        KiteView.__init__(self)
        model.sigSceneModelChanged.connect(self.modelChanged)

    @QtCore.pyqtSlot()
    def modelChanged(self):
        self.main_widget.update()
        self.main_widget.transFromFrame()

        self.param_scene.updateValues()
        self.param_frame.updateValues()
        self.param_meta.updateValues()

        self.dialogTransect.close()

    def togglePolygonMask(self, checked):
        polygon_mask = self.model.scene.polygon_mask
        polygon_mask.set_enabled(checked)


class KiteScenePlot(KitePlot):
    def __init__(self, model):
        self.components_available = {
            "displacement": ["Scene.displacement", lambda sp: sp.scene.displacement],
            "theta": ["Scene.theta", lambda sp: sp.scene.theta],
            "phi": ["Scene.phi", lambda sp: sp.scene.phi],
            "thetaDeg": ["Scene.thetaDeg", lambda sp: sp.scene.thetaDeg],
            "phiDeg": ["Scene.phiDeg", lambda sp: sp.scene.phiDeg],
            "unitE": ["Scene.los.unitE", lambda sp: sp.scene.los.unitE],
            "unitN": ["Scene.los.unitN", lambda sp: sp.scene.los.unitN],
            "unitU": ["Scene.los.unitU", lambda sp: sp.scene.los.unitU],
        }

        if model.scene.displacement_px_var is not None:
            self.components_available["displacement_px_var"] = [
                "Scene.displacement_px_var",
                lambda sp: sp.scene.displacement_px_var,
            ]

        self._component = "displacement"

        KitePlot.__init__(self, model=model, los_arrow=True)

        model.sigFrameChanged.connect(self.onFrameChange)
        model.sigSceneModelChanged.connect(self.update)
        self.loadMaskPolygons()

        # Todo: use activateView
        model.sigSceneChanged.connect(self.update)
        self.model = model

    def roiToVertices(self, roi):
        frame = self.model.scene.frame
        return [
            (h.pos().x() / frame.dE, h.pos().y() / frame.dN) for h in roi.getHandles()
        ]

    def verticesToRoi(self, vertices):
        frame = self.model.scene.frame
        return [
            ((v[0] / frame.cols) * frame.lengthE, (v[1] / frame.rows) * frame.lengthN)
            for v in vertices
        ]

    def loadMaskPolygons(self):
        scene = self.model.scene
        for pid, vertices in scene.polygon_mask.polygons.items():
            self.createMaskPolygon(pid, vertices)

    def newMaskPolygon(self):
        scene = self.model.scene
        frame = self.model.scene.frame

        vertices = np.array([(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)])
        vertices[:, 0] *= frame.cols / 6
        vertices[:, 1] *= frame.rows / 6
        vertices[:, 0] += frame.cols / 2
        vertices[:, 1] += frame.rows / 2

        pid = scene.polygon_mask.add_polygon(vertices.tolist())
        self.createMaskPolygon(pid, vertices)

    def createMaskPolygon(self, pid, vertices):
        roi = PolygonMaskROI(
            self.verticesToRoi(vertices),
            pen=pen_roi,
            movable=False,
            removable=True,
            closed=True,
        )
        roi.pid = pid

        roi.sigRegionChangeFinished.connect(self.updatePolygonMask)
        roi.sigRemoveRequested.connect(self.removePolygonMask)
        self.addItem(roi)

    def updatePolygonMask(self, roi):
        vertices = self.roiToVertices(roi)
        self.model.scene.polygon_mask.update_polygon(roi.pid, vertices)

    def removePolygonMask(self, roi):
        self.model.scene.polygon_mask.remove_polygon(roi.pid)
        self.removeItem(roi)

    def onFrameChange(self):
        self.update()
        self.transFromFrame()


class KiteToolTransect(QtWidgets.QDialog):
    def __init__(self, plot, parent=None):
        QtWidgets.QDialog.__init__(self, parent)

        loadUi(get_resource("transect.ui"), baseinstance=self)

        icon = self.style().standardIcon

        self.closeButton.setIcon(icon(QtWidgets.QStyle.SP_DialogCloseButton))
        self.createButton.setIcon(icon(QtWidgets.QStyle.SP_ArrowUp))
        self.removeButton.setIcon(icon(QtWidgets.QStyle.SP_DialogDiscardButton))

        self.plot = plot
        self.poly_line = None

        self.trans_plot = pg.PlotDataItem(
            antialias=True, fillLevel=0.0, fillBrush=pg.mkBrush(0, 127, 0, 150)
        )

        self.plt_wdgt = pg.PlotWidget()
        self.plt_wdgt.setLabels(bottom={"Distance", " m"}, left="Displacement [m]")

        self.plt_wdgt.showGrid(True, True, alpha=0.5)
        self.plt_wdgt.enableAutoRange()
        self.plt_wdgt.addItem(self.trans_plot)

        self.layoutPlot.addWidget(self.plt_wdgt)
        self.plot.image.sigImageChanged.connect(self.updateTransPlot)
        self.createButton.released.connect(self.addPolyLine)
        self.removeButton.released.connect(self.removePolyLine)

        parent.model.sigConfigChanged.connect(self.close)

    def addPolyLine(self):
        [[xmin, xmax], [ymin, ymax]] = self.plot.viewRange()
        self.poly_line = pg.PolyLineROI(
            positions=[
                (xmin + (xmax - xmin) * 0.4, ymin + (ymax - ymin) * 0.4),
                (xmin + (xmax - xmin) * 0.6, ymin + (ymax - ymin) * 0.6),
            ],
            pen=pg.mkPen("g", width=2),
        )
        self.plot.addItem(self.poly_line)
        self.updateTransPlot()

        self.poly_line.sigRegionChangeFinished.connect(self.updateTransPlot)

    def removePolyLine(self):
        if self.poly_line is None:
            return

        self.plot.removeItem(self.poly_line)
        self.poly_line = None
        self.updateTransPlot()

    def closeEvent(self, event):
        self.removePolyLine()

    def updateTransPlot(self):
        if self.poly_line is None:
            return

        transect = np.ndarray((0))
        length = 0
        for line in self.poly_line.segments:
            transect = np.append(
                transect, line.getArrayRegion(self.plot.image.image, self.plot.image)
            )
            p1, p2 = line.listPoints()
            length += (p2 - p1).length()
        # interpolate over NaNs
        nans, x = np.isnan(transect), lambda z: z.nonzero()[0]
        transect[nans] = np.interp(x(nans), x(~nans), transect[~nans])
        length = np.linspace(0, length, transect.size)

        self.trans_plot.setData(length, transect)
        self.plt_wdgt.setLimits(
            xMin=length.min(),
            xMax=length.max(),
            yMin=transect.min(),
            yMax=transect.max() * 1.1,
        )
        return


class ParamScene(KiteParameterGroup):
    def __init__(self, model, plot, **kwargs):
        kwargs["type"] = "group"
        kwargs["name"] = "Scene"
        self.plot = plot

        self.parameters = {
            "min value": lambda plot: np.nanmin(plot.data),
            "max value": lambda plot: np.nanmax(plot.data),
            "mean value": lambda plot: np.nanmean(plot.data),
        }

        self.plot.image.sigImageChanged.connect(self.updateValues)

        KiteParameterGroup.__init__(self, model=self.plot, model_attr=None, **kwargs)

        def changeComponent(parameter):
            self.plot.component = parameter.value()

        p = {
            "name": "display",
            "values": {
                "displacement": "displacement",
                "theta": "theta",
                "phi": "phi",
                "thetaDeg": "thetaDeg",
                "phiDeg": "phiDeg",
                "los.unitE": "unitE",
                "los.unitN": "unitN",
                "los.unitU": "unitU",
            },
            "value": "displacement",
            "tip": "Change the displayed component of the displacement field.",
        }

        if model.scene.displacement_px_var is not None:
            p["values"]["displacement_px_var"] = "displacement_px_var"

        component = pTypes.ListParameter(**p)
        component.sigValueChanged.connect(changeComponent)
        self.pushChild(component)


class ParamSceneFrame(KiteParameterGroup):
    def __init__(self, model, **kwargs):
        kwargs["type"] = "group"
        kwargs["name"] = ".frame"

        self.parameters = OrderedDict(
            [
                ("cols", None),
                ("rows", None),
                ("dN", None),
                ("dE", None),
                ("spacing", None),
                ("llLat", None),
                ("llLon", None),
                ("llNutm", None),
                ("llEutm", None),
                ("utm_zone", None),
                ("utm_zone_letter", None),
            ]
        )

        model.sigFrameChanged.connect(self.updateValues)

        KiteParameterGroup.__init__(self, model=model, model_attr="frame", **kwargs)


class ParamSceneMeta(KiteParameterGroup):
    def __init__(self, model, **kwargs):
        from datetime import datetime as dt

        kwargs["type"] = "group"
        kwargs["name"] = ".meta"

        def str_to_time(d, fmt="%Y-%m-%d %H:%M:%S"):
            return dt.strftime(dt.fromtimestamp(d), fmt)

        self.parameters = OrderedDict(
            [
                ("time_master", lambda sc: str_to_time(sc.meta.time_master)),
                ("time_slave", lambda sc: str_to_time(sc.meta.time_slave)),
                ("time_separation", lambda sc: "%s" % sc.meta.time_separation),
            ]
        )

        model.sigConfigChanged.connect(self.updateValues)

        KiteParameterGroup.__init__(self, model=model, model_attr="scene", **kwargs)

        def update_meta_info(key, value):
            self.model.scene.meta.__setattr__(key, value)

        p = {
            "name": "scene_title",
            "value": self.model.scene.meta.scene_title,
            "type": "str",
            "tip": "Title of the displacement scene",
        }

        self.scene_title = pTypes.SimpleParameter(**p)
        self.scene_title.sigValueChanged.connect(
            lambda v: update_meta_info("scene_title", v.value())
        )

        p = {"name": "scene_id", "value": self.model.scene.meta.scene_id, "type": "str"}

        self.scene_id = pTypes.SimpleParameter(**p)
        self.scene_id.sigValueChanged.connect(
            lambda v: update_meta_info("scene_id", v.value())
        )

        p = {
            "name": "satellite_name",
            "value": self.model.scene.meta.satellite_name,
            "type": "str",
            "tip": "Name of the satellite",
        }

        self.satellite_name = pTypes.SimpleParameter(**p)
        self.satellite_name.sigValueChanged.connect(
            lambda v: update_meta_info("satellite_name", v.value())
        )

        p = {
            "name": "orbital_node",
            "values": {
                "Ascending": "Ascending",
                "Descending": "Descending",
                "Undefined": "Undefined",
            },
            "value": self.model.scene.meta.orbital_node,
            "tip": "Satellite orbit direction",
        }
        self.orbital_node = pTypes.ListParameter(**p)
        self.orbital_node.sigValueChanged.connect(
            lambda v: update_meta_info("orbital_node", v.value())
        )

        self.pushChild(self.orbital_node)
        self.pushChild(self.satellite_name)
        self.pushChild(self.scene_id)
        self.pushChild(self.scene_title)


class DerampParams(KiteParameterGroup):
    def __init__(self, model, **kwargs):
        scene = model.getScene()
        kwargs["type"] = "group"
        kwargs["name"] = "Scene.detrend"

        KiteParameterGroup.__init__(self, model=model, model_attr="scene", **kwargs)

        p = {
            "name": "demean",
            "type": "bool",
            "value": scene.deramp.config.demean,
            "tip": "subtract mean of displacement",
        }
        self.demean = pTypes.SimpleParameter(**p)

        def toggle_demean(param, checked):
            self.model.getScene().deramp.set_demean(checked)

        self.demean.sigValueChanged.connect(toggle_demean)

        p = {
            "name": "applied",
            "type": "bool",
            "value": scene.deramp.config.applied,
            "tip": "detrend the scene",
        }
        self.applied = pTypes.SimpleParameter(**p)

        def toggle_applied(param, checked):
            self.model.getScene().deramp.set_enabled(checked)

        self.applied.sigValueChanged.connect(toggle_applied)

        self.pushChild(self.applied)
        self.pushChild(self.demean)



#!/bin/python
import numpy as np
from pyrocko.guts import Float, Int, Object
from scipy import ndimage

from kite.meta import ADict, Subject


class SceneProcess(Object):
    def __init__(self, scene, *args, **kwargs):
        Object.__init__(self, *args, **kwargs)
        self.scene = scene
        self.evProcessChanged = Subject()

    def __setattr__(self, attr, value):
        if attr in self.T.propnames:
            self.evProcessChanged.notify()
        Object.__setattr__(self, attr, value)

    def apply(self):
        raise NotImplementedError()

    def remove(self):
        raise NotImplementedError()

    def __del__(self):
        self.evProcessChanged.unsubscribeAll()


class StaticOffset(SceneProcess):
    offset = Float.T(default=0.0, help="Static offset")

    def apply(self):
        self.scene.displacement += self.offset

    def remove(self):
        self.scene.displacement -= self.offset


class Downsample(SceneProcess):
    from scipy import ndimage

    factor = Int.T(help="Downsample factor.")

    def __init__(self, *args, **kwargs):
        SceneProcess.__init__(self, *args, **kwargs)
        sc = self.scene
        self.original = ADict()
        self.original.update(
            {
                "displacement": sc.displacement.copy(),
                "theta": sc.theta.copy(),
                "phi": sc.phi.copy(),
                "frame.dLat": sc.frame.dLat,
                "frame.dLon": sc.frame.dLon,
            }
        )

    def apply(self):
        sc = self.scene
        org = self.original
        factor = self.factor

        sx, sy = sc.displacement.shape
        gx, gy = np.ogrid[0:sx, 0:sy]
        regions = sy / factor * (gx / factor) + gy / factor
        indices = np.arange(regions.max() + 1)

        def block_downsample(arr):
            res = ndimage.mean(arr, labels=regions, index=indices)
            res.shape = (sx / factor, sy / factor)
            return res

        sc.displacement = block_downsample(sc.displacement)
        sc.theta = block_downsample(sc.theta)
        sc.phi = block_downsample(sc.phi)
        sc.frame.dLat = org["frame.dLat"] * self.factor
        sc.frame.dLon = org["frame.dLat"] * self.factor

    def remove(self):
        for prop, value in self.original.items():
            self.scene.__setattr__(prop, value)
        del self.original



#!/usr/bin/python2
import argparse as ap
import logging
import sys

from kite.talpa.talpa import talpa


def main(args=None):
    """
    Talpa app deployed through setuptools
    """
    if args is None:
        args = sys.argv[1:]

    epilog = """Talpa is part of the kite InSAR framework.
More at http://pyrocko.org, http://github.com/pyrocko

BriDGes DFG Project, University of Kiel

 Marius Isken (marius.isken@gfz-potsdam.de)
 Henriette Sudhaus"""
    desc = "Crust deformation modeling"

    parser = ap.ArgumentParser(
        prog="talpa",
        epilog=epilog,
        description=desc,
        parents=[],
        formatter_class=ap.RawTextHelpFormatter,
        prefix_chars="-",
        fromfile_prefix_chars=None,
        argument_default=ap.SUPPRESS,
        conflict_handler="resolve",
        add_help=True,
    )

    parser.add_argument(
        "file",
        type=str,
        help="Load SandboxScene from file (.yml)",
        default=None,
        nargs="?",
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=1,
        help="Verbosity, add multiple to increase verbosity.",
    )

    ns = parser.parse_args(args)

    log_level = logging.WARNING - ns.verbose * 10

    logging.basicConfig()
    stream_handler = logging.root.handlers[0]
    stream_handler.setLevel(level=log_level if log_level > 0 else 0)

    talpa(filename=ns.file)


if __name__ == "__main__":
    main()
